###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Self

from everysk.config import settings
from everysk.core.exceptions import SDKValueError
from everysk.core.fields import ListField, StrField, ChoiceField

from everysk.sdk.engines.cryptography import generate_unique_id
from everysk.sdk.entities.base import BaseEntity, ScriptMetaClass
from everysk.sdk.entities.fields import EntityNameField, EntityDescriptionField, EntityLinkUIDField, EntityWorkspaceField, EntityDateTimeField, EntityTagsField
from everysk.sdk.entities.query import Query
from everysk.sdk.entities.script import Script

###############################################################################
#   File Implementation
###############################################################################
class File(BaseEntity, metaclass=ScriptMetaClass):
    """
    This class represents a file entity object and provides methods to validate and manage the entity's data.

    Attributes:
        script (Script): The script object associated with the file.
        id (StrField): The unique identifier of the file.
        workspace (EntityWorkspaceField): The workspace of the file.
        name (EntityNameField): The name of the file.
        tags (EntityTagsField): The tags of the file.
        description (EntityDescriptionField): The description of the file.
        link_uid (EntityLinkUIDField): The link UID of the file.
        date (EntityDateTimeField): The date associated with the file.
        data (StrField): The file data in base64 format.
        url (StrField): The URL of the file.
        content_type (ChoiceField): The content type of the file.

        version (StrField): The version of the datastore.
        created_on (DateTimeField): The created on date of the datastore.
        updated_on (DateTimeField): The updated on date of the datastore.

    Example usage:
        >>> from everysk.sdk.entities.file.base import File
        >>> file = File()
        >>> file.script = 'my_script'
        >>> file.id = 'file_12345678'
        >>> file.name = 'My File'
        >>> file.tags = ['tag1', 'tag2']
        >>> file.description = 'This is a sample file.'
        >>> file.workspace = 'my_workspace'
        >>> file.date = DateTime.fromisoformat('20220101')
        >>> file.data = 'base64_encoded_data_here'
        >>> file.content_type = 'application/pdf'
        >>> file.url = '/1234567891011211234567890'
        >>> file.create()
        >>> print(file)
        {
            'id': 'file_12345678',
            'script': 'my_script',
            'name': 'My File',
            'description': 'This is a sample file.',
            'tags': ['tag1', 'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': '20220101',
            'data': 'base64_encoded_data_here',
            'content_type': 'application/pdf',
            'url': '/1234567891011211234567890'
            'created': '2021-01-01T00:00:00.000000Z',
            'updated': '2021-01-01T00:00:00.000000Z',
        }
    """
    script: Script
    _orderable_attributes = ListField(default=['date', 'created_on', 'updated_on', 'name'])

    id = StrField(regex=settings.FILE_ID_REGEX, max_size=settings.FILE_ID_MAX_SIZE, required_lazy=True, empty_is_none=True)

    name = EntityNameField()
    description = EntityDescriptionField()
    tags = EntityTagsField()
    link_uid = EntityLinkUIDField()
    workspace = EntityWorkspaceField()

    date = EntityDateTimeField()

    hash = StrField(default=None, max_size=settings.FILE_HASH_LENGTH, required_lazy=True)
    data = StrField(default=None, max_size=settings.FILE_DATA_MAX_SIZE_IN_BASE64, required_lazy=True)
    content_type = ChoiceField(default=None, choices=settings.FILE_CONTENT_TYPES, required_lazy=True)
    url = StrField(required_lazy=True)

    def validate(self) -> bool:
        """
        This method validates the entity object and raises an exception if it is not
        valid. The validation is performed by calling the `validate` method of each field
        of the entity.

        Args:
            self (Self): The entity object to be validated.

        Raises:
            FieldValueError: If the entity object is not valid.
            RequiredFieldError: If a required field is missing.

        Returns:
            bool: True if the entity object is valid, False otherwise.

        Example usage:
            >>> file = File()
            >>> file.validate()
            True
        """
        return self.get_response(self_obj=self)

    @classmethod
    def validate_transient(cls, entity_dict: dict) -> Self:
        """
        Validate the entity properties.

        Args:
            entity_dict (dict): The entity properties.

        Returns:
            BaseEntity: The entity object.

        Example usage:
            >>> entity_dict = {'name': 'My Entity'}
            >>> BaseEntity.validate_transient(entity_dict)
        """
        return cls.get_response(params={'entity_dict': entity_dict})

    def _check_entity_to_query(self) -> bool:
        """
        Check the entity object to query.

        Returns:
            bool: True if the entity object is valid.
        """
        super()._check_entity_to_query()
        if self.url and (self.name or self.tags or self.link_uid):
            raise SDKValueError("Can't filter by URL and Name, Tags or Link UID at the same time")

        return True

    def _mount_query(self, query: Query) -> Query:
        """
        Mount the query object.

        Args:
            query (Query): The query object.

        Returns:
            Query: The query object.
        """
        query = super()._mount_query(query)

        if self.url is not None:
            query = query.where('url', self.url)

        return query

    @staticmethod
    def get_id_prefix() -> str:
        """
        Returns the prefix of the file id field value.

        Returns:
            str: The prefix of the file id field value.

        Example usage:
            >>> File.get_id_prefix()
            'file_'
        """
        return settings.FILE_ID_PREFIX

    @staticmethod
    def generate_url() -> str:
        """
        Generate a unique url for the file.

        Returns:
            str: A unique url for the file.

        Example usage:
            >>> File.generate_url()
            '/1234567891011211234567890'
        """
        return f'/{generate_unique_id()}'

    def to_dict(self, add_class_path: bool = False, recursion: bool = False) -> dict:
        """
        Convert the entity to a JSON-serializable dictionary.
        This method converts the entity object into a dictionary that can be easily
        serialized to JSON.

        Args:
            self (Self): The entity instance to convert.
            with_internals (bool, optional): Whether to include internal parameters. Defaults to True.
            recursion (bool, optional): Whether to include nested entities. Defaults to False.

        Returns:
            dict: A dictionary representation of the File entity.
        Raises:
            NotImplementedError: This method should be implemented in subclasses.
        """
        dct: dict = super().to_dict(add_class_path=add_class_path, recursion=recursion)

        if add_class_path is False:
            dct['url'] = f"{settings.FILE_URL_PATH}{self.url}" if self.url else None

        return dct



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

from everysk.core.number import is_float_convertible
from everysk.core.unittests import TestCase


class NumberTestCase(TestCase):

    def test_is_float_convertible(self):
        self.assertFalse(is_float_convertible(None))
        self.assertFalse(is_float_convertible('invalid'))
        self.assertTrue(is_float_convertible('25.1545'))
        self.assertTrue(is_float_convertible(25.1545))



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import builtins

# This sets the Undefined as a global constant
# This works for all Python scripts and processes, but for some
# autocomplete tools like Pylance, Jedi etc... This does not work,
# so we need to search for "<tool> extends builtins" on Google
# to figure out how to solve the problem
try:
    builtins.Undefined
except AttributeError:
    from everysk.core.undefined import UndefinedType
    builtins.Undefined = UndefinedType()

    # Set UndefinedType to be blocked
    UndefinedType.block = True



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources import Parser
from everysk.core.unittests import TestCase, mock


###############################################################################
#   Parser TestCase Implementations
###############################################################################
class APIParserTestCase(TestCase):

    def test_class_name(self):
        self.assertEqual(Parser.class_name(), 'parser')

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_parser__call_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()

        expected_response = {'success': True}
        mock_api_requestor.post.return_value = expected_response
        mock_create_api_requestor.return_value = mock_api_requestor
        response = Parser._Parser__call_method('test_method', key='value')

        mock_create_api_requestor.assert_called_once()
        mock_api_requestor.post.assert_called_once_with('/parsers/test_method', {'key': 'value'})
        self.assertEqual(response, expected_response)

    def test_parse_method(self):
        with mock.patch('everysk.api.api_resources.parser.Parser._Parser__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected_result'
            result = Parser.parse('test_method', key='value')

        mock_call_method.assert_called_once_with('test_method/parse', key='value')
        self.assertEqual(result, 'expected_result')



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['TestCase', 'mock', 'skip', 'skipUnless', 'SkipTest', 'skipIf']
import difflib
import pprint
from time import time
from typing import Any
from unittest import TestCase as PythonTestCase, mock, skip, skipUnless, SkipTest, skipIf
from unittest.util import _common_shorten_repr
from warnings import warn
from everysk.config import settings
from everysk.core.object import BaseDict


class TestCase(PythonTestCase):

    def _callTestMethod(self, method: callable) -> None:
        # For some tests the time could not pass 1 second, gzip tests for example.
        # So we use mock to fix the time for the tests.
        original_time = time()
        with mock.patch('time.time', return_value=original_time):
            # We could not use super here because the stacklevel would be wrong
            if method() is not None:
                warn(
                    f'It is deprecated to return a value that is not None from a test case ({method})',
                    DeprecationWarning,
                    stacklevel=3
                )

    def assertDictEqual(self, d1: dict | BaseDict, d2: dict | BaseDict, msg: str = None):
        # pylint: disable=protected-access
        self.assertIsInstance(d1, (dict, BaseDict), 'First argument is not a dictionary')
        self.assertIsInstance(d2, (dict, BaseDict), 'Second argument is not a dictionary')

        # We need to ensure that both objects are of the same type to proceed
        if isinstance(d1, dict) and isinstance(d2, BaseDict):
            d1 = type(d2)(**d1)

        elif isinstance(d2, dict) and isinstance(d1, BaseDict):
            d2 = type(d1)(**d2)

        if d1 != d2:
            standardMsg = '%s != %s' % _common_shorten_repr(d1, d2) # pylint: disable=consider-using-f-string, invalid-name
            diff = ('\n' + '\n'.join(difflib.ndiff(
                            pprint.pformat(d1).splitlines(),
                            pprint.pformat(d2).splitlines())))
            standardMsg = self._truncateMessage(standardMsg, diff) # pylint: disable=invalid-name
            self.fail(self._formatMessage(msg, standardMsg))


class EveryskMagicMock(mock.MagicMock): # pylint: disable=too-many-ancestors

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        # To avoid infinite recursion inside serialize.dumps we need to remove this method
        delattr(self, settings.SERIALIZE_CONVERT_METHOD_NAME)

mock.MagicMock = EveryskMagicMock



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from requests import Response
from everysk.core.fields import BoolField, ChoiceField, DictField, StrField
from everysk.core.http import HttpPOSTConnection

###############################################################################
#   Slack Class Implementation
###############################################################################
class Slack(HttpPOSTConnection):
    ## Private attributes
    _color_map = DictField(default={
        'danger': '#a90a05',
        'success': '#138138',
        'warning': '#e9932d'
    }, readonly=True)
    is_json = BoolField(default=True, readonly=True)

    ## Public attributes
    color = ChoiceField(default=None, choices=(None, 'danger', 'success', 'warning'))
    message = StrField(default=None, required=True)
    title = StrField(default=None, required=True)

    def get_payload(self) -> dict:
        """
        Convert all key/value on self to a dict object and apply some conversions.
        """
        return {
            'attachments': [{
                'color': self._color_map.get(self.color, '#000000'),
                'blocks': [
                    {
                        'type': 'header',
                        'text': {
                            'type': 'plain_text',
                            'text': self.title,
                            'emoji': True
                        }
                    },
                    {
                        'type': 'divider'
                    },
                    {
                        'type': 'section',
                        'text': {
                            'type': 'mrkdwn',
                            'text': self.message
                        }
                    },
                    {
                        'type': 'divider'
                    }
                ]
            }]
        }

    def send(self) -> Response:
        """ Sends the message to the Slack Channel """
        return self.get_response()



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from functools import cache
import holidays
from holidays.countries import BR
from everysk.core.datetime import Date


class BRHolidays(BR):
    """
    Subclass of BR representing holidays specific to Brazil and also provides further functionality
    """

    def _populate(self, year):
        """
        Populate holidays specific to Brazil for the given year

        Args:
            year (int): The year for which holidays are to be populated

        Usage:
            >>> from everysk.core.datetime.calendar import BRHolidays
            >>> br_holidays = BRHolidays()
            >>> br_holidays._populate(2022)
            >>> print(br_holidays)
            {
                datetime.date(2022, 1, 1): 'Confraternização Universal',
                datetime.date(2022, 4, 15): 'Sexta-feira Santa',
                ...
            }
        """
        super()._populate(year)

        # We need to check individually for these holidays
        try:
            self.pop_named('Início da Quaresma')
        except KeyError:
            pass

        try:
            self.pop_named('Dia do Servidor Público')
        except KeyError:
            pass

        try:
            self.pop_named('Véspera de Natal')
        except KeyError:
            pass

        try:
            self.pop_named('Véspera de Ano-Novo')
        except KeyError:
            pass

class ANBIMA(BRHolidays):
    pass

class BVMF(BRHolidays):
    """
    Subclass of BRHolidays representing holidays specific to the BVMF calendar.
    """
    def _populate(self, year):
        """
        _summary_

        Args:
            year (int): The year for the holidays to be populated

        Usage:
            Display the holidays for the year of 2022

            >>> from everysk.core.datetime.calendar import BVMF
            >>> bvmf_holidays = BVMF()
            >>> bvmf_holidays._populate(2022)
            >>> print(bvmf_holidays)
            {
                datetime.date(2022, 1, 1): 'Confraternização Universal',
                datetime.date(2022, 4, 15): 'Sexta-feira Santa',
                ...
            }
        """
        super()._populate(year)

        if year < 2022:
            self[Date(year, 1, 25)] = 'Aniversário de São Paulo'
            self[Date(year, 11, 20)] = 'Dia da Consciência Negra'


def get_holidays(calendar: str, years: list = range(2000, 2100)) -> dict:
    """
    Uses the holidays library (https://pypi.org/project/holidays/) to retrieve a list of holidays for a specific country

    It also uses a range of years, if more specification needed

    Args:
        calendar (str): Two digit country symbol.
        years (list, optional): List of int years. Ex: [2021, 2022]. Defaults to [2000, ..., 2099].

    Usage:
        Getting holidays for Brazil (BVMF calendar) for the years 2021 and 2022

        >>> from everysk.core.datetime.calendar import get_holidays
        >>> brazil_holidays = get_holidays('BR', years=[2021, 2022])
        >>> print(brazil_holidays)
        {
            datetime.date(2021, 1, 1) : 'Confraternização Universal',
            datetime.date(2021, 4, 2) : 'Sexta-feira Santa',
            ...
        }
    """
    return _get_holidays(calendar=calendar, years=tuple(years))


@cache
def _get_holidays(calendar: str, years: tuple[int]) -> dict:
    """
    Cacheable version of get_holidays.
    """
    holidays.BVMF = BVMF
    holidays.ANBIMA = ANBIMA
    holidays.BRHolidays = BRHolidays

    # We need to remove some holidays for the BR calendar so we change it
    if calendar == 'BR':
        calendar = 'BRHolidays'

    # Every country has public holidays
    # Brazil has optional holidays as well
    categories = ['public']
    if calendar in {'BRHolidays', 'BVMF', 'ANBIMA'}:
        categories.append('optional')

    return {Date(dt.year, dt.month, dt.day): name for dt, name in holidays.country_holidays(calendar, years=years, categories=categories).items()}



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=unused-import

###########################################
# API RESOURCE
###########################################
from everysk.api.api_resources._tests.api_resource import APIResourceTestCase as EveryskLibAPIResourceTestCase

###########################################
# FILE
###########################################
from everysk.api.api_resources._tests.file import APIFileTestCase as EveryskLibAPIFileTestCase

###########################################
# CUSTOM INDEX
###########################################
from everysk.api.api_resources._tests.custom_index import APICustomIndexTestCase as EveryskLibAPICustomIndexTestCase

###########################################
# PRIVATE SECURITY
###########################################
from everysk.api.api_resources._tests.private_security import APIPrivateSecurityTestCase as EveryskLibAPIPrivateSecurityTestCase

###########################################
# WORKSPACE
###########################################
from everysk.api.api_resources._tests.workspace import APIWorkspaceTestCase as EveryskLibAPIWorkspaceTestCase

###########################################
# REPORT TEMPLATE
###########################################
from everysk.api.api_resources._tests.report_template import APIReportTemplateTestCase as EveryskLibAPIReportTemplateTestCase

###########################################
# PORTFOLIO
###########################################
from everysk.api.api_resources._tests.portfolio import APIPortfolioTestCase as EveryskLibAPIPortfolioTestCase

###########################################
# DATASTORE
###########################################
from everysk.api.api_resources._tests.datastore import APIDatastoreTestCase as EveryskLibAPIDatastoreTestCase

###########################################
# WORKER EXECUTION
###########################################
from everysk.api.api_resources._tests.worker_execution import APIWorkerExecutionTestCase as EveryskLibAPIWorkerExecutionTestCase

###########################################
# PARSER
###########################################
from everysk.api.api_resources._tests.parser import APIParserTestCase as EveryskLibAPIParserTestCase

###########################################
# WORKFLOW EXECUTION
###########################################
from everysk.api.api_resources._tests.workflow_execution import APIWorkflowExecutionTestCase as EveryskLibAPIWorkflowExecutionTestCase

###########################################
# WORKFLOW
###########################################
from everysk.api.api_resources._tests.workflow import APIWorkflowTestCase as EveryskLibAPIWorkflowTestCase

###########################################
# CALCULATION
###########################################
from everysk.api.api_resources._tests.calculation import APICalculationTestCase as EveryskLibAPICalculationTestCase

###########################################
# MARKET DATA
###########################################
from everysk.api.api_resources._tests.market_data import APIMarketDataTestCase as EveryskLibAPIMarketDataTestCase

###########################################
# REPORT
###########################################
from everysk.api.api_resources._tests.report import APIReportTestCase as EveryskLibAPIReportTestCase

###########################################
# RETRIEVABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import RetrievableAPIResourceTestCase as EveryskLibRetrievableAPIResourceTestCase

###########################################
# LISTABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import ListableAPIResourceTestCase as EveryskLibListableAPIResourceTestCase

###########################################
# CREATABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import CreatableAPIResourceTestCase as EveryskLibCreatableAPIResourceTestCase

###########################################
# FILTERABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import FilterableAPIResourceTestCase as EveryskLibFilterableAPIResourceTestCase

###########################################
# UPDATABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import UpdatableAPIResourceTestCase as EveryskLibUpdatableAPIResourceTestCase

###########################################
# DELETABLE API
###########################################
from everysk.api.api_resources._tests.api_resource import DeletableAPIResourceTestCase as EveryskLibDeletableAPIResourceTestCase



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime
from everysk.core.exceptions import RequiredError, FieldValueError
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.datastore.base import Datastore


###############################################################################
#   Datastore TestCase Implementation
###############################################################################
class DatastoreTestCase(TestCase):

    def setUp(self):
        self.sample_data = {
            'id': 'dats_12345678',
            'name': 'My Datastore',
            'description': 'This is a sample datastore.',
            'tags': ['tag1', 'tag2'],
            'link_uid': None,
            'workspace': 'my_workspace',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'level': '1',
            'data': {'key1': 'value1', 'key2': 'value2'},
            'version': 'v1',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9)
        }
        self.datastore = Datastore(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_get_id_prefix(self):
        self.assertEqual(Datastore.get_id_prefix(), settings.DATASTORE_ID_PREFIX)

    def test_validate(self):
        expected_data = compress({'class_name': 'Datastore', 'method_name': 'validate', 'self_obj': self.datastore.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')
        datastore: Datastore = self.datastore.copy()

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            datastore.validate()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_validate_error(self):
        datastore: Datastore = self.datastore.copy()
        datastore.data = DateTime.now()
        with self.assertRaisesRegex(FieldValueError, "Datastore data is not a valid json"):
            datastore.validate()

        datastore.data = None
        with self.assertRaisesRegex(RequiredError, "The data attribute is required"):
            datastore.validate()

    def test_query_load_with_id(self):
        expected_data = compress({'class_name': 'Datastore', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'dats_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Datastore(id='dats_1234567891011211234567890', workspace='SampleWorkspace').load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_query_load(self):
        datastore = Datastore(link_uid='SampleLinkUID', workspace='SampleWorkspace')
        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            datastore.load()

        expected_data = compress({'class_name': 'Query', 'method_name': 'load', 'self_obj': datastore.to_query().to_dict(add_class_path=True), 'params': {'offset': None}}, protocol='gzip', serialize='json')
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core import serialize
from everysk.core.compress import compress, decompress, zip_directory_to_str
from everysk.core.datetime import DateTime, Date
from everysk.core.unittests import TestCase, mock


class CompressTestCase(TestCase):
    @mock.patch.object(serialize, 'dumps')
    def test_compress_serialize_none(self, dumps: mock.MagicMock):
        result = compress('a', serialize=None)
        self.assertEqual(result, b'x\x9cK\x04\x00\x00b\x00b')
        dumps.assert_not_called()

    @mock.patch.object(serialize, 'loads')
    def test_decompress_serialize_none(self, loads: mock.MagicMock):
        result = decompress(b'x\x9cK\x04\x00\x00b\x00b', serialize=None)
        self.assertEqual(result, b'a')
        loads.assert_not_called()

    def test_compress_protocol_invalid(self):
        with self.assertRaisesRegex(ValueError, "Unsupported compression protocol 'invalid'. Use 'zlib' or 'gzip'."):
            compress('string', protocol='invalid', serialize='json')

    def test_decompress_protocol_invalid(self):
        with self.assertRaisesRegex(ValueError, "Unsupported decompression protocol 'invalid'. Use 'zlib' or 'gzip'."):
            decompress('string', protocol='invalid', serialize='json')


class CompressZlibJsonTestCase(TestCase):

    def test_compress(self):
        string = 'aa aa aa aa aa aa aa aa'
        self.assertEqual(compress(string, protocol='zlib', serialize='json'), b'x\x9cSJLT\xc0\x86\x94\x00]\xbd\x075')

    def test_decompress(self):
        string = b'x\x9cSJLT\xc0\x86\x94\x00]\xbd\x075'
        self.assertEqual(decompress(string, protocol='zlib', serialize='json'), 'aa aa aa aa aa aa aa aa')

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='zlib', serialize='json')
        self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='json'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='zlib', serialize='json')
        self.assertListEqual(decompress(obj_compressed, protocol='zlib', serialize='json'), obj)


class CompressZlibPickleTestCase(TestCase):

    def setUp(self) -> None:
        self.string = 'aa aa aa aa aa aa aa aa'
        self.str_repr = b'x\x9c3J64705243K1\xb1L23\xb0HJ6\xb04HJ443H\xb64I6\xb5H521LK2\xb6j`\x99*\xcd\x00\x01=\xe2\x89\x89\n\xd8\xd0\x14=\x00\x08%\x14\x15'

    def test_compress(self):
        self.assertEqual(
            compress(self.string, protocol='zlib', serialize='pickle'),
            self.str_repr
        )

    def test_decompress(self):
        self.assertEqual(
            decompress(self.str_repr, protocol='zlib', serialize='pickle'),
            self.string
        )

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='zlib', serialize='pickle')
        self.assertEqual(decompress(obj_compressed, protocol='zlib', serialize='pickle'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='zlib', serialize='pickle')
        self.assertListEqual(decompress(obj_compressed, protocol='zlib', serialize='pickle'), obj)

###############################################################################
#   File Handling Test Case Implementation
###############################################################################
class FileHandlingTestCase(TestCase):

    @mock.patch('os.walk')
    @mock.patch('zipfile.ZipFile.write')
    def test_zip_directory_to_str_returns_expected_data(self, mock_write: mock.MagicMock, mock_walk: mock.MagicMock):
        expected_base64 = 'UEsFBgAAAAAAAAAAAAAAAAAAAAAAAA=='
        mock_walk.return_value = [
            ['/worker', '_', ['main.py']],
            ['/worker/config', '_', ['utils.py', 'config.json']],
            ['/worker/tests', '_', ['main.py']],
            ['/worker/tests/fixtures', '_', ['data.json']]
        ]

        result = zip_directory_to_str('engine/portfolio/tests', 'zip_root_folder', ['config.json'], ['**/tests*'])
        mock_write.assert_has_calls((
            mock.call('/worker/main.py', 'zip_root_folder/../../../../../worker/main.py'),
            mock.call('/worker/config/utils.py', 'zip_root_folder/../../../../../worker/config/utils.py')
        ))
        self.assertEqual(result, expected_base64)


class CompressGzipJsonTestCase(TestCase):

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='gzip', serialize='json')
        self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='json'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='gzip', serialize='json')
        self.assertListEqual(decompress(obj_compressed, protocol='gzip', serialize='json'), obj)


class CompressGzipPickleTestCase(TestCase):

    def test_undefined(self):
        obj_compressed = compress(Undefined, protocol='gzip', serialize='pickle')
        self.assertEqual(decompress(obj_compressed, protocol='gzip', serialize='pickle'), Undefined)

    def test_complex_obj(self):
        obj = [
            '2024-01-01',
            {'datetime': DateTime.now()},
            {'a': 1, 'b': True},
            'Test',
            1.1,
            ['1', '2', '3', Date.today()],
            Undefined,
            {'undefined': Undefined}
        ]
        obj_compressed = compress(obj, protocol='gzip', serialize='pickle')
        self.assertListEqual(decompress(obj_compressed, protocol='gzip', serialize='pickle'), obj)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.exceptions import FieldValueError
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.portfolio.base import Portfolio
from everysk.sdk.entities.script import Script


###############################################################################
#   Script TestCase Implementation
###############################################################################
class ScriptTestCase(TestCase):
    def setUp(self):
        self.mock_klass = mock.MagicMock()
        self.script = Script(Portfolio)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_script_query_none_user_input(self):
        result = Portfolio.script.fetch(None, 'someVariant', 'someWorkspace')
        self.assertIsNone(result)

    def test_script_query_previous_workers_without_id(self):
        user_input = {"someField": "someValue"}
        result = Portfolio.script.fetch(user_input, 'previousWorkers', 'someWorkspace')
        self.assertIsInstance(result, Portfolio)

    def test_script_query_tag_latest_variant(self):

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            Portfolio.script.fetch({"tags": "someTag"}, 'tagLatest', 'someWorkspace')

        expected_data = compress({'class_name': 'Script', 'method_name': 'fetch', 'self_obj': Portfolio.script.to_dict(add_class_path=True), 'params': {'user_input': {'tags': 'someTag'}, 'variant': 'tagLatest', 'workspace': 'someWorkspace'}}, protocol='gzip', serialize='json')
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_script_query_when_response_is_a_list(self):
        with mock.patch('requests.post') as mock_post, \
            mock.patch('everysk.core.http.compress') as mock_dumps:
            mock_dumps.return_value = 'test'
            mock_post.return_value.content = '[{}, {}]'
            mock_post.return_value.status_code = 200
            result = Portfolio.script.fetch(['tag1', 'tag2'], 'tagList', 'someWorkspace')

        mock_dumps.assert_called_once_with(
            {
                'class_name': 'Script',
                'method_name': 'fetch',
                'self_obj': Portfolio.script,
                'params': {
                    'user_input': ['tag1', 'tag2'],
                    'variant': 'tagList',
                    'workspace': 'someWorkspace'
                }
            },
            protocol='gzip',
            serialize='json',
            use_undefined=True,
            add_class_path=True
        )
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=mock_dumps.return_value
        )
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], Portfolio)
        self.assertIsInstance(result[1], Portfolio)

    def test_persist_transient(self):
        entity = Portfolio(name='test', workspace='someWorkspace')

        with mock.patch('requests.post') as mock_post, \
            mock.patch('everysk.core.http.compress') as mock_compress:
            mock_compress.return_value = 'test'
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.script.persist(entity, 'transient')


        mock_compress.assert_called_once_with(
            {
                'class_name': 'Script',
                'method_name': 'persist',
                'self_obj': Portfolio.script,
                'params': {
                    'entity': entity,
                    'persist': 'transient',
                    'consistency_check': False
                }
            },
            protocol='gzip',
            serialize='json',
            use_undefined=True,
            add_class_path=True
        )
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=mock_compress.return_value
        )
        self.assertIsInstance(result, Portfolio)

    def test_init_(self):
        # pylint: disable=protected-access
        script = Portfolio.script
        self.assertEqual(script._klass, Portfolio)

        script = Script(Portfolio)
        self.assertEqual(script._klass, Portfolio)

        script = Script(_klass=Portfolio)
        self.assertEqual(script._klass, Portfolio)

        script = Script(_klass='Portfolio')
        self.assertEqual(script._klass, Portfolio)

        self.assertRaisesRegex(
            FieldValueError,
            "The _klass value 'Foo' must be a class or a string with the class name",
            Script,
            _klass='Foo'
        )



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import pickle
from everysk.core.exceptions import SigningError
from everysk.core.signing import sign, unsign
from everysk.core.unittests import TestCase, mock


@mock.patch.dict('os.environ', {'EVERYSK_SIGNING_KEY': 'c90d39a0e065'})
class SignTestCase(TestCase):
    maxDiff = None

    def test_string(self):
        data = sign('hello world')
        self.assertEqual(
            data,
            b'101f7c5b76a9e9f683d54cec8a666897472d0f5f:hello world'
        )

    def test_bytes(self):
        data = sign(b'hello world')
        self.assertEqual(
            data,
            b'101f7c5b76a9e9f683d54cec8a666897472d0f5f:hello world'
        )

    def test_bytearray(self):
        data = sign(bytearray([1, 2]))
        self.assertEqual(
            data,
            b'24b910971232f2345db5e7c992e97757d1708534:\x01\x02'
        )

    def test_pickle(self):
        data = sign(pickle.dumps('hello world'))
        self.assertEqual(
            data,
            b'15cfd6c50baeef91929dfc2ab7feccd126db5e00:\x80\x04\x95\x0f\x00\x00\x00\x00\x00\x00\x00\x8c\x0bhello world\x94.'
        )


@mock.patch.dict('os.environ', {'EVERYSK_SIGNING_KEY': 'c90d39a0e065'})
class UnsignTestCase(TestCase):

    def test_string(self):
        data = unsign(b'101f7c5b76a9e9f683d54cec8a666897472d0f5f:hello world')
        self.assertEqual(data, b'hello world')

    def test_bytes(self):
        data = unsign(b'101f7c5b76a9e9f683d54cec8a666897472d0f5f:hello world')
        self.assertEqual(data, b'hello world')

    def test_bytearray(self):
        data = unsign(b'24b910971232f2345db5e7c992e97757d1708534:\x01\x02')
        self.assertEqual(data, b'\x01\x02')

    def test_pickle(self):
        data = unsign(b'15cfd6c50baeef91929dfc2ab7feccd126db5e00:\x80\x04\x95\x0f\x00\x00\x00\x00\x00\x00\x00\x8c\x0bhello world\x94.')
        self.assertEqual(pickle.loads(data), 'hello world')

    def test_invalid_signing_key(self):
        with self.assertRaisesRegex(SigningError, 'Error trying to unsign data.'):
            unsign(b'201f7c5b76a9e9f683d54cec8a666897472d0f5f:hello world')



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# Classes that handle HTTP Connections.

# Url: https://requests.readthedocs.io/en/latest/
#      https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers
#      https://developer.mozilla.org/en-US/docs/Glossary/Quality_values
#
import json
import random
import time
from os.path import dirname
from typing import Any

import requests
from everysk.core.exceptions import HttpError, InvalidArgumentError
from everysk.core.object import BaseObject, BaseObjectConfig
from everysk.core.compress import compress
from everysk.core.serialize import loads
from everysk.core.fields import BoolField, DictField, IntField, ListField, StrField
from everysk.core.log import Logger
from everysk.config import settings

log = Logger(name='everysk-lib-core-http-log')

###############################################################################
#   HttpConnectionConfig Class Implementation
###############################################################################
class HttpConnectionConfig(BaseObjectConfig):
    #
    user_agents_file = StrField()
    # List of user agents to be used in the requests.
    user_agents = ListField()

    # Activate/deactivate the use of the verify flag on HTTP requests.
    # By default this is defined in settings.HTTP_REQUESTS_VERIFY
    # but could be defined in the class configuration too.
    ssl_verify = BoolField(default=settings.HTTP_DEFAULT_SSL_VERIFY)

    # Limit for retries
    retry_limit = IntField(default=settings.HTTP_DEFAULT_RETRY_LIMIT)

    # Times that randrange will use to do the next retry
    retry_end_seconds = IntField(default=settings.HTTP_DEFAULT_RETRY_END_SECONDS)
    retry_start_seconds = IntField(default=settings.HTTP_DEFAULT_RETRY_START_SECONDS)

    def __after_init__(self) -> None:
        # Load the user agents from the file
        if self.user_agents_file is None:
            base_dir = dirname(__file__)
            self.user_agents_file = f'{base_dir}/fixtures/user_agents.json'

        if self.user_agents is None:
            with open(self.user_agents_file, 'r', encoding='utf-8') as fd:
                self.user_agents = json.load(fd)

    def get_ssl_verify(self) -> bool:
        """
        Returns a boolean indicating whether the SSL value
        is set.

        Returns:
            bool: True if the SSL verify is set, False otherwise.

        Example:
            >>> http_connection_config = HttpConnectionConfig()
            >>> http_connection_config.get_ssl_verify()
            True
        """
        return self.ssl_verify if settings.HTTP_REQUESTS_VERIFY is Undefined else settings.HTTP_REQUESTS_VERIFY

    def get_random_agent(self) -> str:
        """
        Return a random user agent from the list of user agents.
        """
        return random.choice(self.user_agents)


###############################################################################
#   HttpConnection Class Implementation
###############################################################################
class HttpConnection(BaseObject):
    """
    Base class to use for HTTP connections, has two attributes:
        - timeout: It's in and represent seconds, defaults to 30.
        - url: It's string and will be the destination.
    """
    class Config(HttpConnectionConfig):
        pass

    ## Private attributes
    _config: Config = None # To autocomplete correctly
    _retry_count = IntField(default=1) # Used to control how many times this connection was retry

    ## Public attributes
    headers = DictField(default=None)
    timeout = IntField(default=settings.HTTP_DEFAULT_TIMEOUT)
    url = StrField(default=None)

    def _clean_response(self, response: requests.models.Response) -> requests.models.Response:
        """
        Checks status_code for response, if status_code is different than 200 throws an exception.

        Args:
            response (requests.models.Response): Http response from server.

        Raises:
            HttpError: If something goes wrong raise exception with status_code and content.
        """
        if getattr(response, 'status_code', settings.HTTP_SUCCESS_STATUS_CODES[0]) not in settings.HTTP_SUCCESS_STATUS_CODES:
            raise HttpError(status_code=response.status_code, msg=response.content)

        return response

    def get_headers(self) -> dict:
        """
        Headers needed to send HTTP methods.
        Below are the most common Headers used by browsers,
        we use them to look less like a Bot and more like a valid access.

        Returns:
            dict: A dictionary containing the headers information.

        Example:
            >>> http_connection = HttpConnection()
            >>> http_connection.get_headers()
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'text/html; charset=UTF-8',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        """
        headers = settings.HTTP_DEFAULT_HEADERS.copy()
        if settings.HTTP_USE_RANDOM_USER_AGENT:
            headers['User-Agent'] = self._config.get_random_agent()

        if self.headers is not None:
            headers.update(self.headers)

        return headers

    def get_url(self) -> str:
        """
        Generate the correct url to fetch data from vendor on POST/GET requests.

        Returns:
            str: Containing the correct URL.
        """
        return self.url

    def message_error_check(self, message: str, status_code: int) -> bool: # pylint: disable=unused-argument
        """
        If this method returns True, the connection will be tried again.

        Args:
            message (str): The error message that occurred on the connection.
            status_code (int): The status code of the response.
        """
        return False

    def _get_response_from_url(self) -> requests.models.Response:
        """
        This will be implemented on child classes to really do the connection.
        """
        return None

    def get_response(self) -> requests.models.Response:
        """
        Try to fetch data from self.get_url and calling self._get_response_from_url for the complete response.
        On HttpError, if self.message_error_check is True we will try connect again for a few more times.
        """
        try:
            response = self._clean_response(self._get_response_from_url())
            # After a success we set the value to 1 again
            self._retry_count = 1
        except Exception as error: # pylint: disable=broad-exception-caught
            # Sometimes it can happen that the server is busy, if this happen the error message must be tested
            # and must return true to enable recursion and we will try again the connection.
            message = str(error).lower()
            status_code = getattr(error, 'status_code', 500)
            if self.message_error_check(message, status_code) and self._retry_count < self._config.retry_limit:
                self._retry_count += 1
                # As we have several processes, we use a random number to avoid collision between them.
                time.sleep(random.randint(self._config.retry_start_seconds, self._config.retry_end_seconds))
                response = self.get_response()
            else:
                raise error

        return response

###############################################################################
#   HttpGETConnection Class Implementation
###############################################################################
class HttpGETConnection(HttpConnection):
    """ Class that implements a interface for HTTP GET connections """
    params = DictField()
    user = StrField()
    password = StrField()

    def get_params(self) -> dict:
        """
        This method is used to make the correct params to pass on GET request.
        These params will be added to the URL with & separating then.
        """
        return self.params

    def _get_response_from_url(self) -> requests.models.Response:
        """
        Try to fetch data from url using GET request.
        Note that any dictionary key whose value is None will not be added to the URL's query string.
        """
        params = {
            'url': self.get_url(),
            'headers': self.get_headers(),
            'params': self.get_params(),
            'verify': self._config.get_ssl_verify(),
            'timeout': self.timeout
        }
        if self.user:
            params['auth'] = (self.user, self.password)

        if settings.HTTP_LOG_RESPONSE:
            dct = params.copy()
            # To remove the password in the logs
            if 'auth' in params:
                dct['auth'] = (params['auth'][0], '***********')

            log.debug('HTTP GET request: %s', dct)

        response = requests.get(**params)

        if settings.HTTP_LOG_RESPONSE:
            dct = {
                'status_code': response.status_code,
                'time': response.elapsed.total_seconds(),
                'headers': response.headers,
                'content': response.content
            }
            log.debug('HTTP GET response: %s', dct)

        return response

###############################################################################
#   HttpPOSTConnection Class Implementation
###############################################################################
class HttpPOSTConnection(HttpConnection):
    """
    Class that implements a interface for HTTP POST connections.
    If self.is_json is True the POST method will be a JSON POST,
    otherwise will be a Form POST Data.
    """
    is_json = BoolField(default=True)
    payload = DictField()

    def get_headers(self) -> dict:
        """
        Headers needed to send HTTP Post methods.

        Returns:
            dict: The headers info for POST requests.

        Example:
            >>> http_post = HttpPostConnection()
            >>> http_post.get_headers()
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'application/json; charset=utf-8',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        """
        headers = super().get_headers()
        if self.is_json:
            headers['Content-Type'] = 'application/json; charset=utf-8'
        else:
            headers['Content-Type'] = 'application/x-www-form-urlencoded'

        return headers

    def get_payload(self) -> dict:
        """
        Make the correct payload body to pass on POST request.

        Returns:
            dict: With all the payload information to send alongside the request.
        """
        return self.payload

    def _get_response_from_url(self) -> requests.models.Response:
        """ Try to get/set data on url using POST request. """
        params = {
            'url': self.get_url(),
            'headers': self.get_headers(),
            'verify': self._config.get_ssl_verify(),
            'timeout': self.timeout
        }
        if self.is_json:
            params['json'] = self.get_payload()
        else:
            params['data'] = self.get_payload()

        if settings.HTTP_LOG_RESPONSE:
            log.debug('HTTP POST request: %s', params)

        response = requests.post(**params)

        if settings.HTTP_LOG_RESPONSE:
            dct = {
                'status_code': response.status_code,
                'time': response.elapsed.total_seconds(),
                'headers': response.headers,
                'content': response.content
            }
            log.debug('HTTP POST response: %s', dct)

        return response

###############################################################################
#   HttpPOSTCompressedConnection Class Implementation
###############################################################################
class HttpPOSTCompressedConnection(HttpPOSTConnection):

    def get_headers(self) -> dict:
        """ Headers needed to send HTTP Post methods. """
        headers = super().get_headers()
        headers['Content-Encoding'] = 'gzip'
        return headers

    def get_payload(self) -> dict:
        """ Make the correct payload body to pass on POST request. """
        return compress(self.payload, protocol='gzip', serialize='json', use_undefined=True, add_class_path=True)

    def get_response(self) -> dict:
        """
        Try to fetch data from self.get_url and calling self._get_response_from_url for the complete response.
        On HttpError, if self.message_error_check is True we will try connect again more 5 times.
        Decompress the response.content
        """
        response = super().get_response()
        return loads(response.content, use_undefined=True, instantiate_object=True)

###############################################################################
#   HttpSDKPOSTConnection Class Implementation
###############################################################################
class HttpSDKPOSTConnection(HttpPOSTCompressedConnection):

    is_json = BoolField(default=False, readonly=True)

    class_name = StrField()
    method_name = StrField()
    self_obj: Any = None
    params = DictField()
    timeout = IntField(default=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT)

    def get_url(self) -> str:
        return f'{settings.EVERYSK_SDK_URL}/{settings.EVERYSK_SDK_VERSION}/{settings.EVERYSK_SDK_ROUTE}'

    def get_payload(self) -> dict:
        """ Make the correct payload body to pass on POST request. """
        self.payload = {
            'class_name': self.class_name,
            'method_name': self.method_name,
            'self_obj': self.self_obj,
            'params': self.params
        }
        return super().get_payload()

    def get_headers(self) -> dict:
        """ Headers needed to send HTTP Post methods. """
        headers = super().get_headers()
        everysk_api_sid = settings.EVERYSK_API_SID
        everysk_api_token = settings.EVERYSK_API_TOKEN

        if not everysk_api_sid:
            raise InvalidArgumentError('Invalid API SID')
        if not everysk_api_token:
            raise InvalidArgumentError('Invalid API TOKEN')

        headers['Authorization'] = f'Bearer {everysk_api_sid}:{everysk_api_token}'

        return headers



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources.calculation import Calculation
from everysk.core.unittests import TestCase, mock

###############################################################################
#   Calculation TestCase Implementation
###############################################################################
class APICalculationTestCase(TestCase):

    def test_refresh_test_case(self):
        params = {'some': 'params'}
        retrieve_params = {'params1', 'params2'}
        calculation = Calculation(retrieve_params, params)
        refreshed_calculation = calculation.refresh()

        self.assertIsInstance(refreshed_calculation, Calculation)

    def test_class_name(self):
        self.assertEqual(Calculation.class_name(), 'calculations')

    def test_class_url(self):
        self.assertEqual(Calculation.class_url(), f'/{Calculation.class_name()}')

    @mock.patch('everysk.api.utils.create_api_requestor')
    def test_calculation__call_method(self, mock_create_api_requestor):
        mock_api_requestor = mock.Mock()

        expected_response = {'success': True}
        mock_api_requestor.post.return_value = expected_response
        mock_create_api_requestor.return_value = mock_api_requestor
        response = Calculation._Calculation__call_method('test_method', key='value')

        mock_create_api_requestor.assert_called_once()
        mock_api_requestor.post.assert_called_once_with('/calculations/test_method', {'key': 'value'})
        self.assertEqual(response, expected_response)

    def test_risk_attribution_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.riskAttribution(key='value')

        mock_call_method.assert_called_once_with('risk_attribution', key='value')
        self.assertEqual(result, 'expected result')

    def test_parametric_risk_attribution_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.parametricRiskAttribution(key='value')

        mock_call_method.assert_called_once_with('parametric_risk_attribution', key='value')
        self.assertEqual(result, 'expected result')

    def test_stress_test_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.stressTest(key='value')

        mock_call_method.assert_called_once_with('stress_test', key='value')
        self.assertEqual(result, 'expected result')

    def test_exposure_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.exposure(key='value')

        mock_call_method.assert_called_once_with('exposure', key='value')
        self.assertEqual(result, 'expected result')

    def test_properties_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.properties(key='value')

        mock_call_method.assert_called_once_with('properties', key='value')
        self.assertEqual(result, 'expected result')

    def test_backtest_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.backtest(key='value')

        mock_call_method.assert_called_once_with('backtest', key='value')
        self.assertEqual(result, 'expected result')

    def test_backtest_statistics_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.backtestStatistics(key='value')

        mock_call_method.assert_called_once_with('backtest_statistics', key='value')
        self.assertEqual(result, 'expected result')

    def test_aggregations_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.aggregations(key='value')

        mock_call_method.assert_called_once_with('aggregations', key='value')
        self.assertEqual(result, 'expected result')

    def test_fundamentals_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.fundamentals(key='value')

        mock_call_method.assert_called_once_with('fundamentals', key='value')
        self.assertEqual(result, 'expected result')

    def test_days_to_unwind_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.daysToUnwind(key='value')

        mock_call_method.assert_called_once_with('days_to_unwind', key='value')
        self.assertEqual(result, 'expected result')

    def test_sensitivity_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.sensitivity(key='value')

        mock_call_method.assert_called_once_with('sensitivity', key='value')
        self.assertEqual(result, 'expected result')

    def test_optimize_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.optimize(key='value')

        mock_call_method.assert_called_once_with('optimize', key='value')
        self.assertEqual(result, 'expected result')

    def test_marginal_tracking_error_method(self):
        with mock.patch('everysk.api.api_resources.calculation.Calculation._Calculation__call_method') as mock_call_method:
            mock_call_method.return_value = 'expected result'
            result = Calculation.marginalTrackingError(key='value')

        mock_call_method.assert_called_once_with('marginal_tracking_error', key='value')
        self.assertEqual(result, 'expected result')



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import StrField


_EVERYSK_OTHER_PRIVATE: str = 'private-attribute'
EVERYSK_TEST_OTHER_NAME = StrField(default='test-other-case', readonly=True)
EVERYSK_TEST_OTHER_FAKE = True
EVERYSK_TEST_OTHER_INT: int = 2
EVERYSK_TEST_OTHER_FIELD_INHERIT = StrField(default='{EVERYSK_TEST_NAME} as {EVERYSK_TEST_FAKE}')
EVERYSK_TEST_OTHER_VAR_INHERIT = '{EVERYSK_TEST_FAKE} as {EVERYSK_TEST_NAME}'



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api import utils
from everysk.api.api_resources.api_resource import APIResource

###############################################################################
#   Calculation Implementation
###############################################################################
class Calculation(APIResource):
    """
    A class representing calculations for financial analysis.

    This class inherits from APIResource and provides methods for executing various financial calculations and analyses.
    """
    def refresh(self): # pylint: disable=arguments-differ
        """
        Refreshes the calculation instance.
        This method refreshes the calculation instance by performing any necessary updates or recalculations.

        Usage:
            >>> from everysk.api.api_resources.calculation import Calculation
            >>> data = {'param1', 'param2'}
            >>> retrieve_params = {'some': 'data'}
            >>> calculation = Calculation(retrieve_params, data)
            >>> calculation.refresh()

        Returns:
            Calculation: The refreshed calculation instance.
        """
        return self

    @classmethod
    def class_name(cls):
        """
        This method returns the class name 'calculations'.

        Usage:
            >>> from everysk.api.api_resources.calculation import Calculation
            >>> Calculation.class_name()
            >>> 'calculations'

        Returns:
            str: 'calculations'
        """
        return 'calculations'

    @classmethod
    def class_url(cls):
        """
        This method returns the class URL in the following format: '/calculations'.

        Usage:
            >>> from everysk.api.api_resources.calculation import Calculation
            >>> url = Calculation.class_url()
            >>> print(url)
            >>> '/calculations'

        Returns:
            str: '/calculations'
        """
        return f'/{cls.class_name()}'

    @classmethod
    def __call_method(cls, method, **kwargs):
        """
        Generic helper method to call various calculation methods on the API. This method acts as a central
        dispatcher that sends requests to different calculation endpoints based on the 'method' argument. The
        specific parameters for each calculation should be passed as keyword arguments, which will vary depending
        on the calculation method being called.

        Args:
            method:
                The calculation method to call. This should match one of the API's available endpoints.

            **kwargs:
                Additional keyword arguments. The contents and requirements of kwargs will depend on the specific API method being invoked.

        Returns:
            dict: The response from the API
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f'{cls.class_url()}/{method}'
        response = api_req.post(url, kwargs)
        return response

    @classmethod
    def riskAttribution(cls, **kwargs): # pylint: disable=invalid-name
        """
        Performs a risk attribution analysis on a specified portfolio or a set of securities. This method calculates and
        attributes the portfolio's risk based on the following parameters

        Args:
            portfolio_id (str, optional):
                ID of an existing portfolio. When passing a portfolio's ID as parameter the calculation will be performed using the given portfolio's securities, date, base currency, and nlv.

            securities (list):
                Array of objects to describe the securities in the portfolio. Each object, represents a security with a unique id, symbol, quantity and label.
                `securities` is not required if `portfolio_id` is being provided, check https://everysk.com/api/docs/#-a-name-securities-a-securities for more info.

            date (str date, optional):
                Date in the following format: `YYYYMMDD`. The date instructs the API to use the market conditions and security prices prevailing on that date.

            base_currency (str, optional):
                3-letter ISO 4217 code for currency, check https://support.everysk.com/hc/en-us/articles/115005366147-Supported-Currencies to see all the available currencies.

            nlv (float, optional):
                The net liquidating value of the portfolio (also known as NAV).

            horizon (int, optional):
                Simulates the behavior of each security via their underlying risk factors. It accepts 1, 5 20, or 60.
                For more information, check https://support.everysk.com/hc/en-us/articles/360016288574-Simulation-Horizon

            sampling (int, optional):
                refers to the collection frequency of historical prices and rates to calculate invariant risk factors. Use 1 for daily sampling and 5 for weekly (non-overlapping) sampling.
                Accepted values are 1 or 5. Click here for different horizon and sampling combinations.

            aggregation (str, optional):
                Computes individual security MCTR's and aggregates them according to a supported criteria.
                Check https://support.everysk.com/hc/en-us/sections/360003156634-Risk-Aggregations for more information

            projection (list, optional):
                User supplied array of securities to be used as a top-down factor model. Maximum number of elements in the projection list is 15

            volatility_half_life (int, optional):
                Half life of volatility information in the following months: 0 (no decay), 2, 6, 12, 24, or 48.

            correlation_half_life (int, optional):
                Half life of correlation information in months: 0 (no decay), 2, 6, 12, 24 or 48.

            risk_measure (str, optional):
                MCTR measures forward-looking portfolio risk properties: vol (annualized portfolio P&L volatility), var (portfolio P&L Value-at-Risk), and cvar (portfolio P&L Conditional Value-at-Risk).

            filter (str, optional):
                Selects portfolio identifiers meeting specific criteria, like isolating fixed income securities. It computes across the entire portfolio but displays results only for the filtered subset
                For additional information, click here: https://everysk.com/api/docs/#-a-name-filters-a-filter-expression

            **kwargs: Additional keyword arguments

        Returns:
            dict: The risk attribution result
        """
        return cls.__call_method('risk_attribution', **kwargs)

    @classmethod
    def parametricRiskAttribution(cls, **kwargs): # pylint: disable=invalid-name
        """
        Calculates the magnitude of a security's reaction to changes in underlying factors, most often in terms of its price to other factors.
        `securities` is not required if `portfolio_id` is being provided.

        Args:
            id (str):
                string used to identify the portfolio

            projection (list):
                User supplied array of securities to be used as a top-down factor model. Maximum number of elements in the projection array is 15.

            portfolio_id (str, optional):
                It is the ID of and existing portfolio. When passing a portfolio's ID as parameter the calculation will be performed using the given portfolio's securities, date, base currency and nlv.

            securities (list):
                Array of objects used to describe the securities in the portfolio. Each object has its own ID, symbol, quantity, and label.

            use_cashflow (boolean, optional):
                For fixed income securities, setting the flag to True maps each cash flow event (like interest or principal payments) to specific points on the interest rate curve.
                If False, it uses Macaulay duration to link bond price to adjacent points on the curve.

            sampling (int, optional):
                refers to the collection frequency of historical prices and rates to calculate invariant risk factors. Use 1 for daily sampling and 5 for weekly (non-overlapping) sampling.
                Accepted values are 1 or 5. Click here for different horizon and sampling combinations.

            confidence (str, optional):
                Determines the confidence level for calculation VaR and CVaR. Values accepted: `1sigma`, `2sigma`, `3sigma`, `85%`, `90%`, `95%`, `97%`, or `99%`.
                For additional information click here: https://everysk.com/api/docs/#-a-name-confidence-a-confidence

            historical_days (str, optional):
                Number of business days used to calculate the covariance for the primitive risk factors: `0 (no decay)`, `2`, `6`, `12`, `24` or `48`.

            exponential_decay (float, optional):
                Factor used in the exponentially weighted moving average (ewma): 0 (no decay); it accepts float values from 0 to 1.

            **kwargs (dict): Additional keyword arguments.

        Returns:
            dict: The parametric risk attribution result.
        """
        return cls.__call_method('parametric_risk_attribution', **kwargs)

    @classmethod
    def stressTest(cls, **kwargs): # pylint: disable=invalid-name
        """
        Stress Test calculates the expected behavior of the portfolio under different scenarios, including extreme events.
        For instance, it takes into account that correlations and volatility tend to increase in periods of market distress.

        Args:
            portfolio_id (str, optional):
                ID of an existing portfolio. When passing a portfolio's ID as parameter the calculation will be performed using the given portfolio's securities, date, base currency, and nlv.

            securities (list):
                Array of objects to describe the securities in the portfolio. Each object, represents a security with a unique id, symbol, quantity and label

            date (str date, optional):
                Date in the following format: `YYYYMMDD`. The date instructs the API to use the market conditions and security prices prevailing on that date.

            base_currency (str, optional):
                3-letter ISO 4217 code for currency.

            nlv (float, optional):
                The net liquidating value of the portfolio (also known as NAV).

            horizon (int, optional):
                Simulates the behavior of each security via their underlying risk factors. It accepts 1, 5 20, or 60.

            sampling (int, optional):
                refers to the collection frequency of historical prices and rates to calculate invariant risk factors. Use 1 for daily sampling and 5 for weekly (non-overlapping) sampling.
                Accepted values are 1 or 5. Click here for different horizon and sampling combinations.

            aggregation (str, optional):
                Computes individual security MCTR's and aggregates them according to a supported criteria.
                Check https://support.everysk.com/hc/en-us/sections/360003156634-Risk-Aggregations for more information

            projection (list, optional):
                User supplied array of securities to be used as a top-down factor model. Maximum number of elements in the projection list is 15

            volatility_half_life (int, optional):
                Half life of volatility information in the following months: 0 (no decay), 2, 6, 12, 24, or 48.

            correlation_half_life (int, optional):
                Half life of correlation information in months: 0 (no decay), 2, 6, 12, 24 or 48.

            shock (str, optional):
                The security being used for the stress test.

            magnitude (file):
                The magnitude of the shock. For more details click here: https://everysk.com/api/docs/#-a-name-magnitude-a-magnitude

            confidence (str, optional):
                Determines the confidence level for calculation VaR and CVaR. Values accepted: `1sigma`, `2sigma`, `3sigma`, `85%`, `90%`, `95%`, `97%`, or `99%`.
                For additional information click here: https://everysk.com/api/docs/#-a-name-confidence-a-confidence

            filter (str, optional):
                Selects portfolio identifiers meeting specific criteria, like isolating fixed income securities. It computes across the entire portfolio but displays results only for the filtered subset
                For additional information, click here: https://everysk.com/api/docs/#-a-name-filters-a-filter-expression

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict: The stress test result.
        """
        return cls.__call_method('stress_test', **kwargs)

    @classmethod
    def exposure(cls, **kwargs):
        """
        Calculates the delta-adjusted notional exposure of each security, converted to the base currency of the portfolio.
        To compute the exposures for a portfolio, make an HTTP POST to your calculation resource URI: `POST /calculations/exposure`

        Args:
            portfolio_id (str, optional):
                ID of an existing portfolio. When passing a portfolio's ID as parameter the calculation will be performed using the given portfolio's securities, date, base currency, and nlv.

            securities (list):
                Array of objects to describe the securities in the portfolio. Each object, represents a security with a unique id, symbol, quantity and label

            date (str date, optional):
                Date in the following format: `YYYYMMDD`. The date instructs the API to use the market conditions and security prices prevailing on that date.

            base_currency (str, optional):
                3-letter ISO 4217 code for currency.

            nlv (float, optional):
                The net liquidating value of the portfolio (also known as NAV).

            sampling (int, optional):
                refers to the collection frequency of historical prices and rates to calculate invariant risk factors. Use 1 for daily sampling and 5 for weekly (non-overlapping) sampling.
                Accepted values are 1 or 5. Click here for different horizon and sampling combinations.

            aggregation (str, optional):
                Computes individual security MCTR's and aggregates them according to a supported criteria.
                Check https://support.everysk.com/hc/en-us/sections/360003156634-Risk-Aggregations for more information

            filter (str, optional):
                Selects portfolio identifiers meeting specific criteria, like isolating fixed income securities. It computes across the entire portfolio but displays results only for the filtered subset
                For additional information, click here: https://everysk.com/api/docs/#-a-name-filters-a-filter-expression

            **kwargs (dict): Additional keyword arguments.

        Returns:
            dict: The exposure calculation result.
        """
        return cls.__call_method('exposure', **kwargs)

    @classmethod
    def properties(cls, **kwargs):
        """
        Calculates the overall portfolio properties with a single API call.
        Sensitivities, exposures and risk are aggregated from individual securities to a portfolio level.

        Args:
            portfolio_id (str, optional):
                ID of an existing portfolio. When passing a portfolio's ID as parameter the calculation will be performed using the given portfolio's securities, date, base currency, and nlv.

            securities (list):
                Array of objects to describe the securities in the portfolio. Each object, represents a security with a unique id, symbol, quantity and label

            date (str date, optional):
                Date in the following format: `YYYYMMDD`. The date instructs the API to use the market conditions and security prices prevailing on that date.

            base_currency (str, optional):
                3-letter ISO 4217 code for currency.

            nlv (float, optional):
                The net liquidating value of the portfolio (also known as NAV).

            horizon (int, optional):
                Simulates the behavior of each security via their underlying risk factors. It accepts 1, 5 20, or 60.

            sampling (int, optional):
                refers to the collection frequency of historical prices and rates to calculate invariant risk factors. Use 1 for daily sampling and 5 for weekly (non-overlapping) sampling.
                Accepted values are 1 or 5. Click here for different horizon and sampling combinations.

            aggregation (str, optional):
                Computes individual security MCTR's and aggregates them according to a supported criteria.
                Check https://support.everysk.com/hc/en-us/sections/360003156634-Risk-Aggregations for more information

            projection (list, optional):
                User supplied array of securities to be used as a top-down factor model. Maximum number of elements in the projection list is 10.

            volatility_half_life (int, optional):
                Half life of volatility information in the following months: 0 (no decay), 2, 6, 12, 24, or 48.

            correlation_half_life (int, optional):
                Half life of correlation information in months: 0 (no decay), 2, 6, 12, 24 or 48.

            confidence (str, optional):
                Determines the confidence level for calculation VaR and CVaR. Values accepted: `1sigma`, `2sigma`, `3sigma`, `85%`, `90%`, `95%`, `97%`, or `99%`.
                For additional information click here: https://everysk.com/api/docs/#-a-name-confidence-a-confidence

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict: The calculation properties.
        """
        return cls.__call_method('properties', **kwargs)

    @classmethod
    def backtest(cls, **kwargs):
        """
        Performs a backtest on the provided portfolio using specified parameters.

        Args:
            user_id (int):
                The ID of the user performing the backtest.

            user_role (str):
                The role of the user, which can influence the backtest computation based on access rights.

            time_zone (str):
                User's time zone, used for aligning the timestamps in the backtest results.

            backtest_date1 (str):
                The start date for the backtest period.

            backtest_date2 (str):
                The end date for the backtest period.

            backtest_periodicity (str):
                The periodicity of the backtest calculations (e.g., daily, weekly, monthly).

            backtest_benchmark_ticker (str):
                The ticker symbol of the benchmark against which to compare the portfolio's performance.

            backtest_rfrate_ticker (str):
                The ticker symbol for the risk-free rate used in the backtest.

            projection_tickers (list):
                A list of tickers used for projections in the backtest.

            output_type (str):
                The type of output expected from the backtest (e.g., full, summary).

            **kwargs (dict):
                Additional keyword arguments that can be passed to the backtest function.

        Usage:
            >>> backtest_results = ClassName.backtest(
            >>> ... user_id=123,
            >>> ... user_role='analyst',
            >>> ... time_zone='EST',
            >>> ... backtest_date1='2020-01-01',
            >>> ... backtest_date2='2020-12-31',
            >>> ... backtest_periodicity='monthly',
            >>> ... backtest_benchmark_ticker='SPY',
            >>> ... backtest_rfrate_ticker='US10Y',
            >>> ... output_type='summary'
            >>> )
            >>> print(backtest_results)

        Returns:
            dict: The backtest result.
        """
        return cls.__call_method('backtest', **kwargs)

    @classmethod
    def backtestStatistics(cls, **kwargs): # pylint: disable=invalid-name
        """
        Retrieves backtest statistics by executing a backtest based on the provided parameters.
        This method internally calls a function to compute backtest statistics and returns a detailed result dict.

        Args:
            user_id (int):
                The ID of the user performing the backtest. This is essential for accessing and
                manipulating the user's data securely.

            user_role (str):
                The role of the user which can influence the level of detail and data accessible in the backtest results.

            time_zone (str):
                The user's time zone, which is crucial for aligning the time-sensitive data in the backtest.

            backtest_date1 (str):
                The starting date for the backtest period, used to delineate the time range for the analysis.

            backtest_date2 (str):
                The ending date for the backtest period, defining the closure of the analysis interval.

            backtest_periodicity (str):
                Specifies the frequency of data points (e.g., daily, weekly, monthly) used in the backtest.

            backtest_benchmark_ticker (str):
                Ticker symbol for the benchmark index against which the portfolio is evaluated.

            backtest_rfrate_ticker (str):
                Ticker symbol for the risk-free rate, used in certain financial metrics and calculations.

            projection_tickers (list):
                List of ticker symbols used for projecting the portfolio's performance.

            output_type (str):
                Determines the format of the output (e.g., 'full', 'summary', 'detailed').

            **kwargs (dict):
                Additional keyword arguments that can influence the backtest computation and its results.

        Usage:
            >>> backtest_results = ClassName.backtestStatistics(
            >>> ... user_id=12345,
            >>> ... user_role='analyst',
            >>> ... time_zone='GMT',
            >>> ... backtest_date1='2021-01-01',
            >>> ... backtest_date2='2021-12-31',
            >>> ... backtest_periodicity='monthly',
            >>> ... backtest_benchmark_ticker='SPY',
            >>> ... backtest_rfrate_ticker='US10Y',
            >>> ... output_type='summary'
            >>> ... )
            >>> print(backtest_results)

        Returns:
            dict: The backtest statistics.
        """
        return cls.__call_method('backtest_statistics', **kwargs)

    @classmethod
    def aggregations(cls, **kwargs):
        """
        Performs financial aggregations on a user's portfolio to generatea summary statistics and insights.
        This method calculates aggregated data points based on the provided portfolio.

        Args:
            user_id (int):
                The ID of the user for whom the aggregation is being performed. Essential for identifying
                the user's data and ensuring data security.

            user_role (str):
                The role of the user, which can define access levels and influence the aggregation results.

            user_settings (dict):
                User-specific settings that can affect the aggregation, such as preferences or configurations.

            portfolio (Portfolio):
                The user's portfolio object on which the aggregation is to be performed.

            output_type (str):
                Specifies the format in which the output is generated, e.g., 'summary', 'detailed'.

            **kwargs (dict):
                Additional keyword arguments that can be used to pass extra parameters necessary for
                the aggregation process.

        Usage:
            >>> aggregation_results = ClassName.aggregations(
            >>> ... user_id=12345,
            >>> ... user_role='manager',
            >>> ... user_settings={'setting1': 'value1', 'setting2': 'value2'},
            >>> ... portfolio=portfolio_object,
            >>> ... output_type='detailed'
            >>> )
            >>> print(aggregation_results)

        Returns:
            dict: The aggregation result.
        """
        return cls.__call_method('aggregations', **kwargs)

    @classmethod
    def fundamentals(cls, **kwargs):
        """
        Retrieves fundamental data.

        Args:
            user_id (int):
                The ID of the user for whom the fundamental data is being retrieved. It ensures the
                personalized and secure access to the necessary data.

            user_role (str):
                The role of the user, which may influence the scope of the retrieved fundamental data.

            user_settings (dict):
                Settings that provide context on user preferences or configurations, impacting
                how fundamental data is calculated and presented.

            portfolio (Portfolio):
                The portfolio object for which fundamentals are being calculated.

            projection_tickers (list):
                List of tickers for which the user seeks to project fundamental data.

            output_type (str):
                Determines the format and detail level of the output, such as 'summary' or 'detailed'.

            **kwargs (dict):
                Additional keyword arguments allowing for further customization of the fundamental data retrieval.

        Usage:
            >>> fundamental_data = ClassName.fundamentals(
            >>> ... user_id=12345,
            >>> ... user_role='analyst',
            >>> ... user_settings={'currency': 'USD', 'fiscal_period': 'Q1'},
            >>> ... portfolio=portfolio_object,
            >>> ... output_type='detailed'
            >>> )
            >>> print(fundamental_data)

        Returns:
            dict: The fundamental data.
        """
        return cls.__call_method('fundamentals', **kwargs)

    @classmethod
    def daysToUnwind(cls, **kwargs): # pylint: disable=invalid-name
        """
        Calculates the number of days required to unwind a position in the portfolio, considering the liquidity of each
        asset. This calculation can help in assessing the market risk associated with the portfolio's current composition.

        Args:
            user_id (int):
                The ID of the user for whom the calculation is performed. Ensures that the computation is
                personalized and secure.

            user_role (str):
                The role of the user, impacting the level of detail and type of data accessible in the results.

            user_settings (dict):
                Settings that influence the computation, such as risk preferences or market assumptions.

            portfolio (Portfolio):
                The portfolio object for which the days to unwind are being calculated.

            days (list):
                A list of days over which the unwind period is considered. Must be provided explicitly.

            liquidity_terms (object, optional):
                An object defining the terms for liquidity calculation, such as market impact.
                If not provided, a default set of terms is used.

            **kwargs (dict): Additional keyword arguments to fine-tune the calculation or provide extra context.

        Usage:
            >>> unwind_data = ClassName.daysToUnwind(
            >>> ... user_id=12345,
            >>> ... user_role='analyst',
            >>> ... user_settings={'market_conditions': 'normal'},
            >>> ... portfolio=portfolio_object,
            >>> ... days=[10, 20, 30],
            >>> ... liquidity_terms={'market_impact': 0.5}
            )
            >>> print(unwind_data)

        Returns:
            dict: The days to unwind calculation result.
        """
        return cls.__call_method('days_to_unwind', **kwargs)

    @classmethod
    def sensitivity(cls, **kwargs):
        """
        Calculates sensitivity. `securities` parameter is not required if `portfolio_id` is being provided.
        Click here for more info https://everysk.com/api/docs/#-a-name-securities-a-securities

        Args:
            portfolio_id (str, optional):
                ID of an existing portfolio. When passing a portfolio's ID as parameter the calculation will be performed using the given portfolio's securities, date, base currency, and nlv.

            securities (list):
                Array of objects to describe the securities in the portfolio. Each object, represents a security with a unique id, symbol, quantity and label

            date (str date, optional):
                Date in the following format: `YYYYMMDD`. The date instructs the API to use the market conditions and security prices prevailing on that date.

            base_currency (str, optional):
                3-letter ISO 4217 code for currency. To see all supported currencies click here => https://support.everysk.com/hc/en-us/articles/115005366147-Supported-Currencies

            sensitivity_type (str, optional):
                The sensitivity type. Check https://everysk.com/api/docs/#sensitivity for the accepted values.

            compute_notional (boolean, optional):
                Determines whether sensitivities should be weighted by the notional exposure or unitized.

            **kwargs (dict): Additional keyword arguments.

        Returns:
            dict: The sensitivity calculation result.
        """
        return cls.__call_method('sensitivity', **kwargs)

    @classmethod
    def optimize(cls, **kwargs):
        """
        Executes an optimization calculation for a given portfolio, applying specified user settings and constraints.

        Args:
            user_id (int):
                The ID of the user for whom the optimization is being performed, ensuring personalized
                and secure processing.

            user_role (str):
                The role of the user, which may affect the optimization process and results.

            user_settings (dict):
                Configuration settings specific to the user that can influence the optimization.

            portfolio (Portfolio):
                The portfolio object to be optimized.

            optimization_model (str):
                The model used for optimization. The default model is specified if not provided.

            optimization_date (str):
                The target date for the optimization. The date affects the data used in the optimization process.

            constraints (object):
                An object defining the constraints to be applied during the optimization.

            to_model_portfolio (bool, optional):
                Flag indicating whether the output should be a model portfolio.

            **kwargs (dict):
                Additional keyword arguments allowing for further customization and specification of the optimization process.

        Usage:
            >>> optimization_result = ClassName.optimize(
            >>> ... user_id=12345,
            >>> ... user_role='analyst',
            >>> ... user_settings={'risk_tolerance': 'medium', 'objective': 'minimize_risk'},
            >>> ... portfolio=portfolio_object,
            >>> ... optimization_model='mean_variance',
            >>> ... optimization_date='2023-01-01',
            >>> ... constraints={'max_weight': 0.1}
            >>> )
            >>> print(optimization_result)

        Returns:
            dict: The optimization result.
        """
        return cls.__call_method('optimize', **kwargs)

    @classmethod
    def marginalTrackingError(cls, **kwargs): # pylint: disable=invalid-name
        """
        Tracking error is the divergence between the price behavior of a position or a portfolio and the price behavior fo a benchmark.

        Args:
            portfolio1 (str):
                The portfolio that is going to be analyzed.

            portfolio2 (str):
                The portfolio that is going the be used as a benchmark to calculate the deviation of differences.

            weekly (boolean, optional):
                The Calculation API will use weekly return to calculate the tracking error.

            EWMA (boolean, optional):
                The default is False, if `True`, the exponentially weighted moving average will be applied to volatilities.

            exponential_decay (float, optional):
                Weighting factor determining the rate at which "older" data enter into the calculation of the `EWMA`.
                It accepts float values from 0 to 1.
                This parameter is only used when `EWMA` is True

            nDays (int, optional):
                Number of historical business days used in calculation when `weekly` is False

            nWeeks (int, optional):
                Number of historical weeks used in calculation when `weekly` is True

            **kwargs (dict): Additional keyword arguments.

        Returns:
            dict: The marginal tracking error calculation result.
        """
        return cls.__call_method('marginal_tracking_error', **kwargs)



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api import get_api_config
from everysk.config import settings
from everysk.core.unittests import TestCase


###############################################################################
#   Implementation
###############################################################################

class TestGetAPIConfig(TestCase):
    def test_full_params_provided(self):
        params = {
            'api_entry': 'https://example.com',
            'api_version': 'v2',
            'api_sid': 'abc',
            'api_token': 'token_example123',
            'verify_ssl_certs': True
        }

        config = get_api_config(params)

        self.assertEqual(config, (
            'https://example.com',
            'v2',
            'abc',
            'token_example123',
            True
        ))

    def test_no_params_provided(self):
        settings.EVERYSK_API_URL = 'https://default.com'
        settings.EVERYSK_API_SID = 'xyz'
        settings.EVERYSK_API_TOKEN = 'default_token'
        settings.EVERYSK_API_VERIFY_SSL_CERTS = False

        config = get_api_config({})

        self.assertEqual(config, (
            'https://default.com',
            'v2',
            'xyz',
            'default_token',
            False
        ))

    def test_partial_params_provided(self):
        params = {
            'api_entry': 'https://partial.com',
            'api_version': 'v3',
        }
        settings.EVERYSK_API_SID = 'partial_sid'
        settings.EVERYSK_API_TOKEN = 'partial_token'
        settings.EVERYSK_API_VERIFY_SSL_CERTS = True

        config = get_api_config(params)

        self.assertEqual(config, (
            'https://partial.com',
            'v3',
            'partial_sid',
            'partial_token',
            True
        ))



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.datetime import DateTime
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.unittests import TestCase, mock

from everysk.sdk.entities.portfolio.base import SecuritiesField, Portfolio
from everysk.sdk.entities.portfolio.securities import Securities


###############################################################################
#   Securities Field TestCase Implementation
###############################################################################
class TestSecuritiesField(TestCase):

    def setUp(self):
        self.securities_field = SecuritiesField()

    def test_clean_value_with_list(self):
        test_value = [{'symbol': 'AAPL', 'name': 'Apple Inc.'}]
        cleaned_value = self.securities_field.clean_value(test_value)

        # Ensure the cleaned value is of type `Securities`
        self.assertIsInstance(cleaned_value, Securities)

    def test_clean_value_without_list(self):
        test_value = 'Some value'
        cleaned_value = self.securities_field.clean_value(test_value)

        # Ensure the cleaned value is not modified
        self.assertEqual(cleaned_value, [test_value])

###############################################################################
#   Portfolio TestCase Implementation
###############################################################################
class TestPortfolio(TestCase):

    def setUp(self):
        self.sample_data = {
            'id': 'port_1234567891011211234567890',
            'workspace': 'SampleWorkspace',
            'name': 'SamplePortfolio',
            'tags': ['tag1', 'tag2'],
            'link_uid': 'ABC',
            'description': 'Description',
            'nlv': 1000.0,
            'base_currency': 'USD',
            'date': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'securities': Securities([{'symbol': 'AAPL'}]),
            'version': '1.0',
            'created_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'updated_on': DateTime(2023, 9, 9, 9, 9, 9, 9),
            'level': 'v1',
            'outstanding_shares': 1000,
            'source_hash': 'XYZ',
            'status': 'OK',
            'portfolio_uid': 'UID123',
            'check_securities': False
        }
        self.portfolio = Portfolio(**self.sample_data)
        self.headers = HttpSDKPOSTConnection().get_headers()
        self.api_url = HttpSDKPOSTConnection().get_url()

    def test_static_methods(self):
        self.assertEqual(Portfolio.get_id_prefix(), settings.PORTFOLIO_ID_PREFIX)

    def test_validate(self):
        portfolio: Portfolio = self.portfolio.copy()
        portfolio.check_securities = True
        portfolio.securities = []
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'validate', 'self_obj': portfolio.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            portfolio.validate()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_check(self):
        entity_dict = self.sample_data
        entity_dict['check_securities'] = True
        entity_dict['securities'] = []

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            result = Portfolio.check(entity_dict=entity_dict)

        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'validate', 'self_obj': result.to_dict(add_class_path=True), 'params': {}}, protocol='gzip', serialize='json')
        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_to_csv(self):
        out_csv = self.portfolio.to_csv()
        self.assertIsInstance(out_csv, str)

    def test_query_load_with_id(self):
        portfolio = Portfolio(id='port_1234567891011211234567890', workspace='SampleWorkspace')
        expected_data = compress({'class_name': 'Portfolio', 'method_name': 'retrieve', 'self_obj': None, 'params': {'entity_id': 'port_1234567891011211234567890'}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            portfolio.load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )

    def test_query_load(self):
        portfolio = Portfolio(link_uid='SampleLinkUID', workspace='SampleWorkspace')
        expected_data = compress({'class_name': 'Query', 'method_name': 'load', 'self_obj': portfolio.to_query().to_dict(add_class_path=True), 'params': {'offset': None}}, protocol='gzip', serialize='json')

        with mock.patch('requests.post') as mock_post:
            mock_post.return_value.content = '{}'
            mock_post.return_value.status_code = 200
            portfolio.load()

        mock_post.assert_called_with(
            url=self.api_url,
            headers=self.headers,
            verify=settings.HTTP_DEFAULT_SSL_VERIFY,
            timeout=settings.EVERYSK_SDK_HTTP_DEFAULT_TIMEOUT,
            data=expected_data
        )



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from typing import Any
import inspect
import requests

from everysk.core.exceptions import HttpError, SDKError
from everysk.core.fields import ListField
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.object import BaseDict, BaseObject
from everysk.core.string import import_from_string

###############################################################################
#   Global Functions
###############################################################################
def handler_input_args(input_args: Any) -> Any:
    """
    This function handles the input arguments and returns the dictionary.

    Args:
        input_args (Any): The input arguments to be handled.

    Returns:
        Any: The parsed object.
    """
    ret: Any = input_args
    if isinstance(input_args, dict):
        ret = BaseDict()
        for key, value in input_args.items():
            ret[key] = handler_input_args(value)

    elif isinstance(input_args, list):
        ret = type(input_args)()
        for item in input_args:
            ret.append(handler_input_args(item))

    return ret

###############################################################################
#   BaseSDK Class Implementation
###############################################################################
class BaseSDK(BaseObject):
    """
    A base class for SDK classes.

    This class provides a base implementation for SDK classes.
    """
    # This is a blacklist to not create keys
    _keys_blacklist = ListField(default=['query', 'script'], readonly=True)

    @classmethod
    def get_response(cls, **kwargs: dict) -> Any:
        """
        Get a response from an SDK method.

        This method sends an HTTP POST request to a remote service and returns the response as a dictionary.

        Args:
            **kwargs (dict): Keyword arguments used to configure the HTTP request and SDK behavior.

        Keyword Args:
            class_name (str, optional): The name of the SDK class making the request.
                Defaults to the class name of the calling class.
            method_name (str, optional): The name of the SDK method making the request.
                Defaults to the name of the calling function.
            self_obj (object, optional): An instance of the calling SDK class, if applicable. Defaults to None.
            params (dict, optional): Additional parameters to include in the HTTP request. Defaults to an empty dictionary.

        Returns:
            Any: The response from the remote service.

        Raises:
            SDKError: If there is an issue with the SDK operation.
        """

        # Set default values for keyword arguments if not provided
        kwargs.setdefault('class_name', cls.__name__)
        kwargs.setdefault('method_name', inspect.stack()[1].function)
        kwargs.setdefault('self_obj', None)
        kwargs.setdefault('params', {})

        try:
            response: requests.models.Response = HttpSDKPOSTConnection(**kwargs).get_response()
        except HttpError as error:
            raise SDKError(error.msg) from error

        if response and isinstance(response, (dict, BaseDict)) and 'error_message' in response and 'error_module' in response:
            error_module = import_from_string(response['error_module'])
            raise error_module(response['error_message'])

        return response



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
##############################################################################
__all__ = ['compress', 'decompress']
import gzip
import zlib
import io
import os
import zipfile
import base64
import fnmatch
from typing import Any
from everysk.core.log import Logger
from everysk.core.serialize import dumps, loads

###############################################################################
#   Globals
###############################################################################

log = Logger('everysk-lib-compress')

###############################################################################
#   Public Functions Implementation
###############################################################################

def compress_json(obj: Any) -> str:
    """
    This function first serializes the input object into a JSON string using the dumps function.
    Then it compresses the JSON string using zlib compression.
    The output is a bytes object which may contain non-printable characters due to compression

    Args:
        obj (Any):
            The JSON-serializable object to compress

    Usage:
        >>> data = {'key': 'value'}
        >>> compressed_data = compress_json(data)
        >>> print(compressed_data)
        >>> b'x\x9c\xcbH\xcd\xc9\xc9W(....)' # Example of the compressed data

    Returns:
        str: The compressed string representation of the JSON-serialized object
    """
    log.deprecated("compress_json is deprecated. Use compress(obj, serialize='json') instead.")
    return compress(obj, serialize='json')

def decompress_json(data: str, convert_str_to_date: bool = False) -> Any: # pylint: disable=unused-argument
    """
    Decompress data with zlib and transform to an obj with loads function.
    The input data should be a zlib-compressed string.
    The returned object may be any valid JSON-serializable Python object

    Args:
        data (str): the zlib-compressed string to decompress and deserialized data
        convert_str_to_date (bool): Enable conversion of str to Date and Datetime. Default is True
    Usage:
        >>> from everysk.core.compress import decompress_json
        >>> compressed_data = b'x\x9c\xabV\xcaN\xadT\xb2RP*K\xcc)MU\xaa\x05\x00+\xaf\x05A'
        >>> decompressed_data = decompress_json(compressed_data)
        >>> print(decompressed_data)
        >>> {'key': 'value'} # Example of the decompressed data'

    Returns:
        Any: The Python object reconstructed from the decompressed and deserialized
    """
    log.deprecated("decompress_json is deprecated. Use decompress(obj, serialize='json') instead.")
    return decompress(data, serialize='json')

def compress_pickle(obj: Any) -> str:
    """
    Convert obj to string with pickle dumps then uses zlib to compress it.
    The output is a bytes object which may contain non-printable characters due to compression.

    Args:
        obj (Any): The Python object to compress

    Example:
        >>> data = {'key': 'value'}
        >>> compressed_data = compress_pickle(data)
        >>> print(compressed_data)
        >>> b'x\x9c\xabV*I,.Q(...)'  # Example compressed string output

    Returns:
        str: The compressed string representation of the serialized object.
    """
    log.deprecated("compress_pickle is deprecated. Use compress(obj, serialize='pickle') instead.")
    return compress(obj, serialize='pickle')

def decompress_pickle(data: str) -> Any:
    """
    Decompress data with zlib and transform to a obj with pickle loads.
    The input data should be a zlib-compressed string generated from a pickled Python object.

    Args:
        data (str): The zlib-compressed string to decompress and deserialize.

    Example:
        >>> compressed_data = b'x\x9c\xabV\xcaN\xadT\xb2RP*K\xcc)MU\xaa\x05\x00+\xaf\x05A' # Example compressed data
        >>> decompressed_data = decompress_pickle(compressed_data)
        >>> print(decompressed_data)
        >>> {'key': 'value'}  # Example decompressed object output

    Returns:
        Any:
            The Python object reconstructed from the decompressed and deserialized data.
    """
    log.deprecated("decompress_pickle is deprecated. Use decompress(obj, serialize='pickle') instead.")
    return decompress(data, serialize='pickle')

def compress(obj: Any, protocol: str = 'zlib', serialize: str | None = 'pickle', use_undefined: bool | None = True, add_class_path: bool | None = None) -> bytes:
    """
    Compress an object using either zlib or gzip compression.
    If serialize is set, the object is serialized using the specified serialization format before compression.
    Supported options for serialization are JSON or Pickle.

    Args:
        obj (Any): The Python object to compress
        protocol (str): The serialization protocol to use. Default is 'zlib'. Options are 'zlib' and 'gzip'.
        serialize (str): The serialization format to use. Default is 'pickle'. Options are 'json', 'pickle' and None.
        use_undefined (bool): If True, undefined values are included in the serialization. Default is True.
        add_class_path (bool): If True, the class path is included in the serialization. Default is None.

    Example:
        >>> data = {'key': 'value'}
        >>> compress(data)
        >>> b'x\x9c\xabV*I,.Q(...)'  # Example compressed string output

    Returns:
        bytes: The compressed string representation of the serialized object.
    """
    if protocol == 'zlib':
        compress_fn = zlib.compress
    elif protocol == 'gzip':
        compress_fn = gzip.compress
    else:
        raise ValueError(f"Unsupported compression protocol '{protocol}'. Use 'zlib' or 'gzip'.")

    result = obj
    if serialize:
        result = dumps(obj, protocol=serialize, use_undefined=use_undefined, add_class_path=add_class_path)

    if isinstance(result, str):
        result = result.encode('utf-8')

    return compress_fn(result)

def decompress(data: bytes, protocol: str = 'zlib', serialize: str | None = 'pickle', use_undefined: bool | None = True, instantiate_object: bool | None = True) -> Any:
    """
    Decompress data using either zlib or gzip compression.
    If serialize is set, the result from decompress is serialized using the specified serialization converting it to a python object.
    Supported options for serialization are JSON or Pickle.

    Args:
        data (bytes): The compressed data to decompress.
        protocol (str): The serialization protocol to use. Default is 'zlib'. Options are 'zlib' and 'gzip'.
        serialize (str): The serialization format to use. Default is 'pickle'. Options are 'json', 'pickle' and None.
        use_undefined (bool): If True, undefined values are included in the serialization. Default is True.
        instantiate_object (bool): If True, the object is instantiated. Default is True.

    Example:
        >>> compressed_data = b'x\x9c\xabV\xcaN\xadT\xb2RP*K\xcc)MU\xaa\x05\x00+\xaf\x05A' # Example of compressed data
        >>> decompress(compressed_data)
        >>> {'key': 'value'}

    Returns:
        Any: If serialize is set returns a Python object otherwise a bytes object.
    """
    if protocol == 'zlib':
        decompress_fn = zlib.decompress
    elif protocol == 'gzip':
        decompress_fn = gzip.decompress
    else:
        raise ValueError(f"Unsupported decompression protocol '{protocol}'. Use 'zlib' or 'gzip'.")

    result = decompress_fn(data)
    if serialize:
        result = loads(result, protocol=serialize, use_undefined=use_undefined, instantiate_object=instantiate_object)

    return result

def zip_directory_to_str(path: str, path_name: str, ignore_files: list = None, ignore_roots: list = None) -> str:
    """
    This function takes a directory path as input and creates a zip file in memory.

    Args:
        path (str): The path to the directory to zip.
        path_name (str): The name of the root folder inside the zip file.
        ignore_files (list): A list of file names to ignore when zipping the directory.
        ignore_roots (list): A list of patterns to ignore when zipping the directory.

    Example:
        >>> zip_directory_to_str('path/to/directory', 'root_folder_name')
        >>> zip_directory_to_str('path/to/directory', 'root_folder_name', ['config.json'], ['**/log', '**/temp/**'])

    Returns:
        str: The base64 encoded string representation of the zip file.
    """
    # Create a BytesIO object to store the zip file
    memory_zip = io.BytesIO()

    # Create a ZipFile object with the memory buffer as its destination
    with zipfile.ZipFile(memory_zip, 'w', zipfile.ZIP_DEFLATED) as zip_file:
        for root, _, files in os.walk(path):
            # Check if the current root matches any of the ignore_roots patterns
            if any(fnmatch.fnmatch(root, pattern) for pattern in ignore_roots):
                continue
            for file in files:
                # Ignore files if specified
                if ignore_files and file in ignore_files:
                    continue
                file_path = os.path.join(root, file)
                # Compute the relative path and add the path_name as the root folder
                relative_path = os.path.relpath(file_path, path)
                zip_file.write(file_path, os.path.join(path_name, relative_path))

    # We need to adjust the cursor of the BytesIO object to the start after writing
    memory_zip.seek(0)

    # Return the base64 encoded string representation of the zip file
    return base64.b64encode(memory_zip.getvalue()).decode('utf-8')



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
import json
from os import environ
from everysk.config import settings
from everysk.core.compress import compress
from everysk.core.exceptions import HttpError, InvalidArgumentError
from everysk.core.http import HttpConnection, HttpConnectionConfig, HttpGETConnection, HttpPOSTConnection, HttpSDKPOSTConnection, HttpPOSTCompressedConnection, requests, time,log
from everysk.core.object import BaseObjectConfig
from everysk.core.unittests import TestCase, mock


class FakeResponse:
    # https://everysk.atlassian.net/browse/COD-1047
    def __bool__(self):
        return False


class HttpConnectionConfigTestCase(TestCase):

    def test_default_attributes(self):
        obj = HttpConnectionConfig()
        self.assertEqual(obj.retry_end_seconds, 30)
        self.assertEqual(obj.retry_limit, 5)
        self.assertEqual(obj.retry_start_seconds, 5)
        self.assertIsInstance(HttpConnection._config, HttpConnectionConfig)

    def test_get_ssl_verify(self):
        obj = HttpConnectionConfig()
        self.assertTrue(obj.get_ssl_verify())
        obj.ssl_verify = False
        self.assertFalse(obj.get_ssl_verify())

    def test_get_ssl_verify_environment(self):
        obj = HttpConnectionConfig()
        self.assertTrue(obj.get_ssl_verify())
        environ['HTTP_REQUESTS_VERIFY'] = 'False'
        self.assertFalse(obj.get_ssl_verify())
        obj.ssl_verify = False
        environ['HTTP_REQUESTS_VERIFY'] = 'True'
        self.assertTrue(obj.get_ssl_verify())
        environ.pop('HTTP_REQUESTS_VERIFY')

    def test_user_agent_list(self):
        obj = HttpConnectionConfig()
        with open(obj.user_agents_file, 'r', encoding='utf-8') as fd:
            user_agents = json.load(fd)

        self.assertListEqual(obj.user_agents, user_agents)

    def test_get_random_agent(self):
        obj = HttpConnectionConfig()
        self.assertIn(obj.get_random_agent(), obj.user_agents)
        self.assertNotEqual(obj.get_random_agent(), obj.get_random_agent())

    def test_inheritance(self):
        self.assertTrue(issubclass(HttpConnectionConfig, BaseObjectConfig))


class HttpConnectionTestCase(TestCase):

    def test_get_url(self):
        http = HttpConnection(url='https://test.com')
        self.assertEqual(http.url, 'https://test.com')
        self.assertEqual(http.get_url(), 'https://test.com')

    def test_clean_response_200(self):
        response = mock.MagicMock()
        response.status_code = 200
        http = HttpConnection()
        self.assertEqual(http._clean_response(response), response) # pylint: disable=protected-access

    def test_clean_response_201(self):
        response = mock.MagicMock()
        response.status_code = 201
        http = HttpConnection()
        self.assertEqual(http._clean_response(response), response) # pylint: disable=protected-access

    def test_clean_response_202(self):
        response = mock.MagicMock()
        response.status_code = 200
        http = HttpConnection()
        self.assertEqual(http._clean_response(response), response) # pylint: disable=protected-access

    def test_clean_response_303(self):
        response = mock.MagicMock()
        response.status_code = 303
        http = HttpConnection()
        self.assertEqual(http._clean_response(response), response) # pylint: disable=protected-access

    def test_clean_response_500(self):
        response = mock.MagicMock()
        response.status_code = 500
        http = HttpConnection()
        self.assertRaises(HttpError, http._clean_response, response) # pylint: disable=protected-access

    def test_clean_without_status_code(self):
        # https://everysk.atlassian.net/browse/COD-1047
        response = FakeResponse()
        self.assertFalse(response)
        http = HttpConnection()
        http._clean_response(response) # pylint: disable=protected-access

    def test_get_headers(self):
        http = HttpConnection()
        self.assertDictEqual(
            http.get_headers(),
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'text/html; charset=UTF-8',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        )

    def test_get_headers_attribute(self):
        http = HttpConnection(headers={'key': 'value'})
        self.assertDictEqual(
            http.get_headers(),
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'text/html; charset=UTF-8',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
                'key': 'value'
            }
        )

    def test_get_random_agent(self):
        obj = HttpConnection()
        self.assertEqual(obj.get_headers()['User-Agent'], obj.get_headers()['User-Agent'])
        with mock.patch('os.environ', {'HTTP_USE_RANDOM_USER_AGENT': 1}):
            self.assertNotEqual(obj.get_headers()['User-Agent'], obj.get_headers()['User-Agent'])

    def test_message_error_check(self):
        self.assertFalse(HttpConnection().message_error_check('text', 500))

    def test_get_response_from_url(self):
        self.assertIsNone(HttpConnection()._get_response_from_url()) # pylint: disable=protected-access

    @mock.patch.object(HttpConnection, '_clean_response', return_value='response')
    def test_get_response(self, _clean_response: mock.MagicMock):
        self.assertEqual(HttpConnection().get_response(), 'response')
        _clean_response.assert_called_once_with(None)

    @mock.patch.object(HttpConnection, '_clean_response', side_effect=HttpError('Error'))
    @mock.patch.object(HttpConnection, 'message_error_check', return_value=True)
    @mock.patch.object(time, 'sleep')
    def test_get_response_error(
        self, sleep: mock.MagicMock, message_error_check: mock.MagicMock, _clean_response: mock.MagicMock
    ):
        self.assertRaises(HttpError, HttpConnection().get_response)
        self.assertEqual(sleep.call_count, 4)
        self.assertEqual(message_error_check.call_count, 5)
        self.assertEqual(_clean_response.call_count, 5)


class HttpGETConnectionTestCase(TestCase):

    def setUp(self) -> None:
        self.old_get = requests.get
        self.response = mock.MagicMock()
        self.response.status_code = 200
        requests.get = mock.create_autospec(requests.get)
        requests.get.return_value = self.response

    def tearDown(self) -> None:
        requests.get = self.old_get

    def test_get_params(self):
        http = HttpGETConnection(url='https://test.com')
        self.assertIsNone(http.get_params())
        http.params = {'key': 'value'}
        self.assertDictEqual(http.get_params(), {'key': 'value'})

    def test_get_response(self):
        http = HttpGETConnection(url='https://test.com', params={'p1': 1, 'p2': 2})
        response = http.get_response()
        requests.get.assert_called_once_with(
            url='https://test.com',
            headers=http.get_headers(),
            params={'p1': 1, 'p2': 2},
            verify=True,
            timeout=30
        )
        self.assertEqual(response, requests.get.return_value)

    def test_user(self):
        http = HttpGETConnection(
            url='https://test.com',
            params={'p1': 1, 'p2': 2},
            user='user',
            password='pass'
        )
        response = http.get_response()
        requests.get.assert_called_once_with(
            url='https://test.com',
            headers=http.get_headers(),
            params={'p1': 1, 'p2': 2},
            timeout=30,
            verify=True,
            auth=('user', 'pass')
        )
        self.assertEqual(response, requests.get.return_value)

    @mock.patch.object(log, 'debug')
    def test_http_log_response(self, debug: mock.MagicMock):
        http = HttpGETConnection(url='https://test.com', params={'p1': 1, 'p2': 2}, user='user', password='pass')
        settings.HTTP_LOG_RESPONSE = True
        response = http.get_response()
        settings.HTTP_LOG_RESPONSE = False
        debug.assert_has_calls([
            mock.call('HTTP GET request: %s', {'url': 'https://test.com', 'headers': http.get_headers(), 'params': {'p1': 1, 'p2': 2}, 'verify': True, 'timeout': 30, 'auth': ('user', '***********')}),
            mock.call('HTTP GET response: %s', {'status_code': 200, 'time': self.response.elapsed.total_seconds.return_value, 'headers': self.response.headers, 'content': self.response.content})
        ])
        self.assertEqual(response, requests.get.return_value)

class HttpPOSTConnectionTestCase(TestCase):

    def setUp(self) -> None:
        self.old_post = requests.post
        self.response = mock.MagicMock()
        self.response.status_code = 200
        requests.post = mock.create_autospec(requests.post)
        requests.post.return_value = self.response

    def tearDown(self) -> None:
        requests.post = self.old_post

    def test_get_headers(self):
        http = HttpPOSTConnection(url='https://test.com')
        self.assertDictEqual(
            http.get_headers(),
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'application/json; charset=utf-8',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        )

    def test_get_payload(self):
        http = HttpPOSTConnection(url='https://test.com')
        self.assertIsNone(http.get_payload())
        http.payload = {'key': 'value'}
        self.assertDictEqual(http.get_payload(), {'key': 'value'})

    def test_get_response(self):
        http = HttpPOSTConnection(url='https://test.com', payload={'p1': 1, 'p2': 2})
        response = http.get_response()
        requests.post.assert_called_once_with(
            headers=http.get_headers(),
            json={'p1': 1, 'p2': 2},
            url='https://test.com',
            verify=True,
            timeout=30
        )
        self.assertEqual(response, requests.post.return_value)

    def test_post_not_json(self):
        http = HttpPOSTConnection(url='https://test.com', is_json=False, payload={'p1': 1, 'p2': 2})
        response = http.get_response()
        self.assertDictEqual(
            http.get_headers(),
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'application/x-www-form-urlencoded',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        )
        requests.post.assert_called_once_with(
            headers=http.get_headers(),
            data={'p1': 1, 'p2': 2},
            url='https://test.com',
            verify=True,
            timeout=30
        )

        self.assertEqual(response, requests.post.return_value)

    @mock.patch.object(log, 'debug')
    def test_http_log_response(self, debug: mock.MagicMock):
        http = HttpPOSTConnection(url='https://test.com', is_json=False, payload={'p1': 1, 'p2': 2})
        settings.HTTP_LOG_RESPONSE = True
        response = http.get_response()
        settings.HTTP_LOG_RESPONSE = False
        debug.assert_has_calls([
            mock.call('HTTP POST request: %s', {'url': 'https://test.com', 'headers': http.get_headers(), 'data': {'p1': 1, 'p2': 2}, 'verify': True, 'timeout': 30}),
            mock.call('HTTP POST response: %s', {'status_code': 200, 'time': self.response.elapsed.total_seconds.return_value, 'headers': self.response.headers, 'content': self.response.content})
        ])
        self.assertEqual(response, requests.post.return_value)

class HttpSDKPOSTConnectionTestCase(TestCase):

    def setUp(self) -> None:
        self.old_post = requests.post
        response = mock.MagicMock()
        response.status_code = 200
        requests.post = mock.create_autospec(requests.post)
        requests.post.return_value = response
        settings.EVERYSK_API_SID = 'SID'
        settings.EVERYSK_API_TOKEN = 'TOKEN'
        self.class_name = 'class_name'
        self.method_name = 'method_name'
        self.self_obj = 'self_obj'
        self.param = 'params'

    def tearDown(self) -> None:
        requests.post = self.old_post

    def test_init_without_url(self):
        http = HttpSDKPOSTConnection()
        self.assertIsNone(http.url)
        self.assertEqual(http.get_url(), 'https://api.everysk.com/v1/sdk_function')

    def test_get_headers_without_SID(self):
        settings.EVERYSK_API_SID = None
        http = HttpSDKPOSTConnection()
        with self.assertRaisesRegex(InvalidArgumentError, 'Invalid API SID'):
            http.get_headers()

    def test_get_headers_without_token(self):
        settings.EVERYSK_API_TOKEN = None
        http = HttpSDKPOSTConnection()
        with self.assertRaisesRegex(InvalidArgumentError, 'Invalid API TOKEN'):
            http.get_headers()

    def test_get_headers(self):
        http = HttpSDKPOSTConnection(url='https://test.com')
        self.assertDictEqual(
            http.get_headers(),
            {
                'Accept-Encoding': 'gzip, deflate;q=0.9',
                'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
                'Authorization': 'Bearer SID:TOKEN',
                'Cache-control': 'no-cache',
                'Connection': 'close',
                'Content-Type': 'application/x-www-form-urlencoded',
                'Content-Encoding': 'gzip',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36'
            }
        )

    def test_get_payload(self):
        http = HttpSDKPOSTConnection(url='https://test.com', class_name='class_name', method_name='method_name', self_obj={'self_obj': 'obj'}, params={})
        self.assertEqual(http.get_payload(), compress({'class_name': 'class_name', 'method_name': 'method_name', 'self_obj': {'self_obj': 'obj'}, 'params': {}}, protocol='gzip', serialize='json'))

    def test_get_url(self):
        http = HttpSDKPOSTConnection()
        expected_url = f'{settings.EVERYSK_SDK_URL}/{settings.EVERYSK_SDK_VERSION}/{settings.EVERYSK_SDK_ROUTE}'
        ret = http.get_url()

        self.assertEqual(ret, expected_url)

class HttpPOSTCompressedConnectionTestCase(TestCase):

    def test_get_payload(self):
        http = HttpPOSTCompressedConnection(url='https://test.com', payload={'p1': 1, 'p2': 2})
        self.assertEqual(http.get_payload(),  compress({'p1': 1, 'p2': 2}, protocol='gzip', serialize='json'))



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
from everysk.api.api_resources.api_resource import (
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource
)
from everysk.api import utils

###############################################################################
#   Worker Execution Implementation
###############################################################################
class WorkerExecution(
    RetrievableAPIResource,
    ListableAPIResource,
    DeletableAPIResource
):

    @classmethod
    def retrieve(cls, **kwargs):
        """
        Retrieves details of a single worker execution or a list of worker executions based on the provided arguments.
        This class method sends a GET request to the API to retrieve a worker execution.

        Args:
            worker_execution_id (str, optional):
                The ID of the worker execution to retrieve. If specified, the method
                returns details for this specific worker execution. If omitted, the
                method may return a list of worker executions based on other filters.

            workflow_execution_id (str, optional):
                The ID of the workflow execution to fetch the worker executions for.
                This argument is considered only if 'worker_execution_id' is not provided.

            with_result (bool, optional):
                Indicates whether to include detailed results in the response. The default
                behavior depends on the context of the request (detailed for a single execution,
                summary for a list).

        Usage:
            # Retrieve details for a specific worker execution
            >>> worker_execution = WorkerExecution.retrieve(worker_execution_id='12345')

            # Retrieve a list of worker executions for a specific workflow execution
            >>> worker_executions = WorkerExecution.retrieve(workflow_execution_id='67890')

        Returns:
            WorkerExecution: An instance of the WorkerExecution class
        """
        api_req = utils.create_api_requestor(kwargs)
        url = f'/workflows{cls.class_url()}'
        response = api_req.get(url, kwargs)
        return utils.to_object(WorkerExecution, kwargs, response)



###############################################################################
#
# (C) Copyright 2024 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
__all__ = ['dumps', 'loads']
import json
import pickle
from typing import Any
from everysk.config import settings
from everysk.core.object import BaseObject, CLASS_KEY
from everysk.core.datetime import DateTime, Date
from everysk.core.undefined import UndefinedType
from everysk.core.signing import sign, unsign
from everysk.core.string import import_from_string

try:
    from simplejson import JSONEncoder as _JSONEncoder # type: ignore
except ImportError:
    from json import JSONEncoder as _JSONEncoder


###############################################################################
#   Constants
###############################################################################
DATE_KEY: str = '__date__'
DATETIME_KEY: str = '__datetime__'
PICKLE_KEY_SEPARATOR: str = ':'
UNDEFINED_KEY: str = '__undefined__'


###############################################################################
#   JSON Encoder Class Implementation
###############################################################################
class JSONEncoder(_JSONEncoder):

    ## Public attributes
    add_class_path: bool = True
    date_format: str = None
    datetime_format: str = None
    use_undefined: bool = False

    def __init__(self, **kwargs) -> None:
        # Set specific params
        self.add_class_path = kwargs.pop('add_class_path', self.add_class_path)
        self.date_format = kwargs.pop('date_format', self.date_format)
        self.datetime_format = kwargs.pop('datetime_format', self.datetime_format)
        self.use_undefined = kwargs.pop('use_undefined', self.use_undefined)

        # Set all default params
        super().__init__(**kwargs)

    def undefined_to_dict(self, obj) -> dict[str, None] | None: # pylint: disable=unused-argument
        """
        Convert an Undefined object to a string or None.
        If `self.use_undefined` is set to True, it returns the default parse string of the Undefined object.

        Args:
            obj (UndefinedType): The Undefined object to convert to a string.
        """
        if self.use_undefined:
            return {UNDEFINED_KEY: None}

        return None

    def convert_date(self, obj: Date) -> dict[str, str] | str:
        """
        Convert a Date object to a dict or string.
        If `self.date_format` is set the result will be a string in this format
        otherwise the result will be a dict with the Date as isoformat.

        Args:
            obj (Date): The Date object to convert to a string or dict.
        """
        if self.date_format:
            return obj.strftime(self.date_format)

        return {DATE_KEY: obj.isoformat()}

    def convert_datetime(self, obj: DateTime) -> dict[str, str] | str:
        """
        Convert a DateTime object to a dict or string.
        If `self.datetime_format` is set the result will be a string in this format
        otherwise the result will be a dict with the DateTime as isoformat.

        Args:
            obj (DateTime): The DateTime object to convert to a string.
        """
        if self.datetime_format:
            return obj.strftime(self.datetime_format)

        return {DATETIME_KEY: obj.isoformat()}

    def convert_base_object(self, obj: Any) -> Any:
        """
        At this point we've translated the model instance into Python native datatypes.
        To finalize the serialization process we render the data into json.

        Args:
            obj (Any): The object to convert to a JSON serializable format.

        returns:
            Any: The JSON serializable
        """
        func = getattr(obj, settings.SERIALIZE_CONVERT_METHOD_NAME)
        return func(add_class_path=self.add_class_path, recursion=False)

    def default(self, obj: Any) -> dict | None: # pylint: disable=arguments-renamed
        """
        Convert an object to a JSON serializable format.

        Args:
            obj (Any): The object to convert to a JSON serializable format.
        """
        if Date.is_date(obj):
            return self.convert_date(obj)

        if DateTime.is_datetime(obj):
            return self.convert_datetime(obj)

        if hasattr(obj, settings.SERIALIZE_CONVERT_METHOD_NAME):
            return self.convert_base_object(obj)

        if isinstance(obj, (frozenset, set)):
            return list(obj)

        if obj is Undefined:
            return self.undefined_to_dict(obj)

        return super().default(obj)


###############################################################################
#   JSON Decoder Class Implementation
###############################################################################
class JSONDecoder(json.JSONDecoder):

    ## Public attributes
    add_class_path: bool = True
    date_format: str = None
    datetime_format: str = None
    instantiate_object: bool = True
    use_undefined: bool = False

    def __init__(self, **kwargs) -> None:
        # Set specific params
        self.add_class_path = kwargs.pop('add_class_path', self.add_class_path)
        self.date_format = kwargs.pop('date_format', self.date_format)
        self.datetime_format = kwargs.pop('datetime_format', self.datetime_format)
        self.instantiate_object = kwargs.pop('instantiate_object', self.instantiate_object)
        self.use_undefined = kwargs.pop('use_undefined', self.use_undefined)

        if 'object_hook' not in kwargs or kwargs['object_hook'] is None:
            kwargs['object_hook'] = self.custom_object_hook

        # Set all default params
        super().__init__(**kwargs)

    def str_to_date(self, obj: str) -> Date:
        """
        Convert a string to a Date object.
        If `self.date_format` is set, it uses the specified date format to convert the string to a Date object.

        Args:
            obj (str): The string to convert to a Date object.
        """
        if self.date_format:
            return Date.strptime(obj, self.date_format)

        return Date.fromisoformat(obj)

    def str_to_datetime(self, obj: str) -> DateTime:
        """
        Convert a string to a DateTime object.
        If `self.datetime_format` is set, it uses the specified datetime format to convert the string to a DateTime object.

        Args:
            obj (str): The string to convert to a DateTime object.
        """
        if self.datetime_format:
            return DateTime.strptime(obj, self.datetime_format)

        return DateTime.fromisoformat(obj)

    def str_to_undefined(self, obj: str) -> UndefinedType | None: # pylint: disable=unused-argument
        """
        Convert a string to Undefined object, if `self.use_undefined` is set to True, otherwise return None.

        Args:
            obj (str): The string to convert to a Undefined object.
        """
        if self.use_undefined:
            return Undefined

        return None

    def import_from_string(self, path: str, obj: dict) -> BaseObject:
        """
        Import a class from a string path.

        Args:
            path (str): The path to the class to import.
        """
        try:
            klass = import_from_string(path)
        except ImportError:
            klass = BaseObject

        return klass(**obj)

    def _get_correct_path(self, path: str) -> str:
        """
        Get the correct path for the class.

        Args:
            path (str): The path to the class to import.
        """
        klass_name: str = path.split('.')[-1]
        if klass_name in settings.EVERYSK_SDK_ENTITIES_MODULES_PATH:
            return settings.EVERYSK_SDK_ENTITIES_MODULES_PATH[klass_name]
        if klass_name in settings.EVERYSK_SDK_ENGINES_MODULES_PATH:
            return settings.EVERYSK_SDK_ENGINES_MODULES_PATH[klass_name]
        if klass_name in settings.EVERYSK_SDK_MODULES_PATH:
            return settings.EVERYSK_SDK_MODULES_PATH[klass_name]

        return path

    def dict_to_base_obj_class(self, obj: dict) -> BaseObject | dict:
        """
        Convert a dictionary to a instance of BaseObject or return
        the dictionary if the flag instantiate_object is False.

        Args:
            obj (dict): A dictionary.

        Returns:
            BaseObject: The BaseObject instance.
        """
        path: str = obj.pop(CLASS_KEY)
        if not self.instantiate_object:
            return obj

        # For SDK classes we need to change the path to avoid problems inside the Client project
        if path.startswith('everysk.sdk.'):
            path = self._get_correct_path(path)

        return self.import_from_string(path, obj)


    def custom_object_hook(self, obj: dict) -> UndefinedType | None | Date | DateTime:
        """
        We change the default object hook to handle custom object hooks for date, datetime and Undefined objects.

        Args:
            obj (dict): A dictionary object to convert to check and convert.
        """
        if DATE_KEY in obj:
            return self.str_to_date(obj[DATE_KEY])

        if DATETIME_KEY in obj:
            return self.str_to_datetime(obj[DATETIME_KEY])

        if CLASS_KEY in obj:
            return self.dict_to_base_obj_class(obj)

        if UNDEFINED_KEY in obj:
            return self.str_to_undefined(obj[UNDEFINED_KEY])

        return obj


###############################################################################
#   Global Functions Implementation
###############################################################################
def dumps(
    obj: Any,
    *, # Limits that only named arguments can be passed after this
    allow_nan: bool = True,
    check_circular: bool = True,
    cls: type = JSONEncoder,
    date_format: str = None,
    datetime_format: str = None,
    default: callable = None,
    ensure_ascii: bool = True,
    indent: int = None,
    protocol: str = 'json',
    separators: tuple = None,
    skipkeys: bool = False,
    sort_keys: bool = False,
    use_undefined: bool = False,
    add_class_path: bool = True,
    **kwargs
) -> str | bytes:
    """
    Serialize `obj` to a JSON/Pickle formatted `str`.

    If `allow_nan` is false, then it will be a `ValueError` to
    serialize out of range `float` values (`nan`, `inf`, `-inf`) in
    strict compliance of the JSON specification, instead of using the
    JavaScript equivalents (`NaN`, `Infinity`, `-Infinity`).

    If `check_circular` is false, then the circular reference check
    for container types will be skipped and a circular reference will
    result in an `RecursionError` (or worse).

    The date_format and datetime_format parameters can be used to specify the
    date and datetime formats to use when serializing date and datetime objects.
    If not specified, the default ISO format is used.

    If `ensure_ascii` is false, then the return value can contain non-ASCII
    characters if they appear in strings contained in `obj`. Otherwise, all
    such characters are escaped in JSON strings.

    If `indent` is a non-negative integer, then JSON array elements and
    object members will be pretty-printed with that indent level. An indent
    level of 0 will only insert newlines. `None` is the most compact
    representation.

    The protocol argument defines the encoding protocol to use. By default, it is 'json'
    and at the moment we only support json and pickle.

    If specified, `separators` should be an `(item_separator, key_separator)`
    tuple.  The default is `(', ', ': ')` if *indent* is `None` and
    `(',', ': ')` otherwise.  To get the most compact JSON representation,
    you should specify `(',', ':')` to eliminate whitespace.

    `default(obj)` is a function that should return a serializable version
    of obj or raise TypeError. The default simply raises TypeError.

    If *sort_keys* is true (default: `False`), then the output of
    dictionaries will be sorted by key.

    To use a custom `JSONEncoder` subclass (e.g. one that overrides the
    `.default()` method to serialize additional types), specify it with
    the `cls` kwarg; otherwise `JSONEncoder` is used.

    If `skipkeys` is true then `dict` keys that are not basic types
    (`str`, `int`, `float`, `bool`, `None`) will be skipped
    instead of raising a `TypeError`.

    The `use_undefined` parameter can be used to serialize `Undefined` objects
    as a string. If set to True, the default parse string of the `Undefined` object is used.
    Otherwise, `Undefined` objects are serialized as `None`.
    """
    if protocol == 'json':
        if isinstance(obj, bytes):
            obj = obj.decode(json.detect_encoding(obj))

        return json.dumps(
            obj, allow_nan=allow_nan, check_circular=check_circular, cls=cls, date_format=date_format,
            datetime_format=datetime_format, default=default, ensure_ascii=ensure_ascii, indent=indent,
            separators=separators, skipkeys=skipkeys, sort_keys=sort_keys, use_undefined=use_undefined,
            add_class_path=add_class_path, **kwargs
        )
    if protocol == 'pickle':
        # Generate the Pickle data
        result = pickle.dumps(obj)
        if settings.EVERYSK_SIGNING_KEY:
            return sign(result)

        return result

    raise ValueError(f"Unsupported serialize protocol '{protocol}'. Use 'json' or 'pickle'.")

def loads(
    data: str | bytes | bytearray,
    *, # Limits that only named arguments can be passed after this
    cls: type = JSONDecoder,
    date_format: str | None = None,
    datetime_format: str | None = None,
    object_hook: callable = None,
    object_pairs_hook: callable = None,
    parse_constant: callable = None,
    parse_float: callable = None,
    parse_int: callable = None,
    protocol: str = 'json',
    use_undefined: bool = False,
    instantiate_object: bool = True,
    **kwargs
) -> Any:
    """
    Deserialize ``data`` (a ``str``, ``bytes`` or ``bytearray`` instance
    containing a JSON/Pickle document) to a Python object.

    ``object_hook`` is an optional function that will be called with the
    result of any object literal decode (a ``dict``). The return value of
    ``object_hook`` will be used instead of the ``dict``. This feature
    can be used to implement custom decoders (e.g. JSON-RPC class hinting).

    ``object_pairs_hook`` is an optional function that will be called with the
    result of any object literal decoded with an ordered list of pairs.  The
    return value of ``object_pairs_hook`` will be used instead of the ``dict``.
    This feature can be used to implement custom decoders.  If ``object_hook``
    is also defined, the ``object_pairs_hook`` takes priority.

    ``parse_float``, if specified, will be called with the string
    of every JSON float to be decoded. By default this is equivalent to
    float(num_str). This can be used to use another datatype or parser
    for JSON floats (e.g. decimal.Decimal).

    ``parse_int``, if specified, will be called with the string
    of every JSON int to be decoded. By default this is equivalent to
    int(num_str). This can be used to use another datatype or parser
    for JSON integers (e.g. float).

    ``parse_constant``, if specified, will be called with one of the
    following strings: -Infinity, Infinity, NaN.
    This can be used to raise an exception if invalid JSON numbers
    are encountered.

    The protocol argument defines the encoding protocol to use. By default, it is 'json'
    and at the moment we only support json and pickle.

    To use a custom ``JSONDecoder`` subclass, specify it with the ``cls``
    kwarg; otherwise ``JSONDecoder`` is used.

    ``instantiate_object`` is an optional flag that can be used to
    return a dictionary without the ``class_path`` key. This can be useful when we don't
    want to instantiate the object.
    """
    if protocol == 'json':
        if isinstance(data, bytes):
            data = data.decode(json.detect_encoding(data))

        return json.loads(
            data, cls=cls, date_format=date_format, datetime_format=datetime_format, object_hook=object_hook,
            object_pairs_hook=object_pairs_hook, parse_constant=parse_constant, parse_float=parse_float,
            parse_int=parse_int, use_undefined=use_undefined, instantiate_object=instantiate_object,
            **kwargs
        )

    if protocol == 'pickle':
        if settings.EVERYSK_SIGNING_KEY:
            data = unsign(data)

        return pickle.loads(data)

    raise ValueError(f"Unsupported serialize protocol '{protocol}'. Use 'json' or 'pickle'.")



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################

###############################################################################
#   Imports
###############################################################################
import os
import requests

from everysk.core.serialize import dumps
from everysk.core.compress import compress
from everysk.core.object import BaseDict
from everysk.core.exceptions import HttpError, SDKError
from everysk.core.http import HttpSDKPOSTConnection
from everysk.core.unittests import TestCase, mock

from everysk.sdk.base import BaseSDK, handler_input_args
from everysk.sdk.entities.portfolio.base import Portfolio

###############################################################################
#   BaseSDK TestCase Implementation
###############################################################################
class TestBaseSDK(TestCase):
# pylint: disable=no-name-in-module
# pylint: disable=import-outside-toplevel
# pylint: disable=unused-import

    def setUp(self):
        self.default_kwargs = {
            'class_name': 'BaseSDK',
            'method_name': 'setUp',  # This will be overridden in each test method.
            'self_obj': None,
            'params': {}
        }
        self.old_post = requests.post
        self.headers = {
            'Accept-Encoding': 'gzip, deflate;q=0.9',
            'Accept-Language': 'en-US, en;q=0.9, pt-BR;q=0.8, pt;q=0.7',
            'Cache-control': 'no-cache',
            'Connection': 'close',
            'Content-Type': 'application/x-www-form-urlencoded',
            'Content-Encoding': 'gzip',
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36',
            'Authorization': 'Bearer SID:TOKEN'
        }
        os.environ['EVERYSK_API_URL_SCHEME'] = 'https'
        os.environ['EVERYSK_API_URL_DOMAIN'] = 'test.com'
        os.environ['EVERYSK_API_SID'] = 'SID'
        os.environ['EVERYSK_API_TOKEN'] = 'TOKEN'
        response = mock.MagicMock()
        response.content = dumps({'my_SID': 'teste'})
        response.status_code = 200
        requests.post = mock.create_autospec(requests.post)
        requests.post.return_value = response

    def tearDown(self) -> None:
        requests.post = self.old_post

    def test_import_from_sdk_path(self):
        from everysk.sdk.entities import Portfolio as root_Portfolio

        self.assertEqual(root_Portfolio, Portfolio)

        with self.assertRaisesRegex(ImportError, "cannot import name 'Module' from 'everysk.sdk.entities'"):
            from everysk.sdk.entities import Module

    def test_defaults_are_set(self):
        # Using the context manager to mock HttpSDKPOSTConnection
        http = BaseSDK.get_response(**self.default_kwargs)

        requests.post.assert_called_once_with(
            url='https://test.com/v1/sdk_function',
            headers=self.headers,
            verify=True,
            timeout=3600,
            data=compress(self.default_kwargs, protocol='gzip', serialize='json')
        )

    def test_automatic_class_name_assignment(self):
        # In this scenario, we are not passing 'class_name'.
        # Directly calling get_response from BaseSDK
        kwargs = {'method_name': 'test_automatic_class_name_assignment'}
        BaseSDK.get_response(**kwargs)

        expected_kwargs = {**self.default_kwargs, **kwargs, 'class_name': 'BaseSDK'}

        requests.post.assert_called_once_with(
            url='https://test.com/v1/sdk_function',
            headers=self.headers,
            verify=True,
            timeout=3600,
            data=compress(expected_kwargs, protocol='gzip', serialize='json')
        )

    def test_automatic_method_name_assignment(self):
        # In this scenario, we are not passing 'class_name'.
        # Directly calling get_response from BaseSDK
        kwargs = {'class_name': 'BaseSDK'}
        BaseSDK.get_response(**kwargs)

        expected_kwargs = {**self.default_kwargs, **kwargs, 'method_name': 'test_automatic_method_name_assignment'}

        requests.post.assert_called_once_with(
            url='https://test.com/v1/sdk_function',
            headers=self.headers,
            verify=True,
            timeout=3600,
            data=compress(expected_kwargs, protocol='gzip', serialize='json')
        )

    def test_get_response_raises_sdk_error_on_http_error(self):
        with mock.patch.object(HttpSDKPOSTConnection, 'get_response', side_effect=HttpError("Test HTTP Error")):
            with self.assertRaises(SDKError) as context:
                BaseSDK.get_response()
            self.assertEqual(str(context.exception), "Test HTTP Error")

    def test_get_response_with_sdk_error(self):
        with mock.patch.object(HttpSDKPOSTConnection, 'get_response', return_value={'error_module':'everysk.core.exceptions.SDKError', 'error_message':'SDK sample error'}):
            with self.assertRaisesRegex(SDKError, 'SDK sample error'):
                BaseSDK.get_response()

    ###############################################################################
    #   handler input args TestCase Implementation
    ###############################################################################
    def test_args_handler_with_no_args(self):
        input_dct = dict()
        expected = BaseDict()
        res = handler_input_args(input_dct)
        self.assertDictEqual(res, expected)
        self.assertIsInstance(res, BaseDict)

    def test_args_handler_with_args(self):
        input_dct = {'a': 1, 'b': 2}
        expected = BaseDict(**input_dct)
        res = handler_input_args(input_dct)
        self.assertDictEqual(res, expected)
        self.assertIsInstance(res, BaseDict)

    def test_args_handler_with_non_dict(self):
        self.assertEqual(handler_input_args(None), None)

    def test_args_handler_with_complex_dict(self):
        input_dct = {'a': 1, 'b': {'c': 2}}
        expected = BaseDict(a=1, b=BaseDict(c=2))
        res = handler_input_args(input_dct)

        self.assertDictEqual(res, expected)
        self.assertIsInstance(res, BaseDict)

    def test_args_handler_with_complex_dict_and_list(self):
        input_dct = {'a': 1, 'b': {'c': [{'a': [{'a': 'b'}]}, 3]}}
        expected = BaseDict(a=1, b=BaseDict(c=[BaseDict(a=[BaseDict(a='b')]), 3]))
        res = handler_input_args(input_dct)
        self.assertDictEqual(res, expected)
        self.assertIsInstance(handler_input_args(res), BaseDict)
        self.assertIsInstance(res['b'], BaseDict)
        self.assertIsInstance(res['b']['c'][0], BaseDict)
        self.assertIsInstance(res['b']['c'][0]['a'][0], BaseDict)

    def test_args_handler_with_string(self):
        expected = '123'
        res = handler_input_args(expected)
        self.assertEqual(expected, res)
        self.assertNotIsInstance(res, BaseDict)

    def test_arg_handler_with_integer(self):
        expected = 123
        res = handler_input_args(expected)
        self.assertEqual(expected, res)
        self.assertNotIsInstance(res, BaseDict)

    def test_arg_handler_with_float(self):
        expected = 3.14
        res = handler_input_args(expected)
        self.assertEqual(expected, res)
        self.assertNotIsInstance(res, BaseDict)
        self.assertIsInstance(res, float)

    def test_arg_handler_with_complex_list(self):
        input_args = [42, 3.14, 'a', ['b', 1], [{'d': 2, 'e': 6, 'list': [1, 2, 3, 4, 5]}]]
        expected = [42, 3.14, 'a', ['b', 1], [BaseDict(d=2, e=6, list=[1, 2, 3, 4, 5])]]
        res = handler_input_args(input_args)
        self.assertIsInstance(res, list)
        self.assertIsInstance(res[3], list)
        self.assertIsInstance(res[4][0]['list'], list)
        self.assertEqual(expected, res)
